from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "unplugged.services.admin"
    verbose_name = "Admin Service"
    label = "services_admin"

    def ready(self):
        from .handler import AdminService  # NOQA



import logging
import threading
from collections import defaultdict

import wrapt
from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db import models
from jsonfield import JSONField
from marshmallow import INCLUDE

from ..baseplugin import PluginBase
from ..pluginhandler import pluginhandler
from ..signals import plugin_loaded, plugin_unloaded

logger = logging.getLogger(__name__)


class PluginCache:
    def __init__(self):
        self.clear()

    def clear(self):
        self.plugins = defaultdict(dict)
        self.plugins_list = []

    def __contains__(self, key):
        if isinstance(key, tuple):
            plugin_type, name = key
        else:
            plugin_type = key.plugin_type
            name = key.name

        if name in self.plugins[plugin_type]:
            return True
        else:
            return False

    def get_plugin_by_keys(self, plugin_type, name):
        return self.plugins[plugin_type][name]

    def get_plugin(self, plugin):
        return self.plugins[plugin.plugin_type][plugin.name]

    def add_plugin(self, plugin):
        self.plugins_list.append(plugin)
        self.plugins[plugin.plugin_type][plugin.name] = plugin

    def remove_plugin(self, plugin):
        self.plugins_list.remove(plugin)
        del self.plugins[plugin.plugin_type][plugin.name]


PLUGIN_CACHE = PluginCache()
PLUGIN_CREATE_LOCK = threading.Lock()
PLUGIN_CREATE_LOCKS = {}


class PluginProxy(wrapt.ObjectProxy):
    @property
    def __wrapped__(self):
        plugin_type, name = self.__getattrib__("__wrapped_info__")
        key = (plugin_type, name)
        if key in PLUGIN_CACHE:
            return PLUGIN_CACHE.get_plugin_by_keys(plugin_type, name)
        else:
            return None

    @__wrapped__.setter
    def __wrapped__(self, value):
        self.__setattrib__("__wrapped_info__", value.plugin_type, value.name)


class PluginManager(models.Manager):
    def bootstrap(self):
        logger.info("Bootstrapping plugin manager")
        for plugin_type in settings.PLUGIN_INITIALIZATION_ORDER:
            Plugin.objects.initialize_plugins(plugin_type)

        # from ..simpleadmin import scan_and_register_simpleadmin_template

        # scan_and_register_simpleadmin_template()

    def initialize_plugins(self, plugin_type):
        logger.info(f"Initializing plugins of type {plugin_type}")
        for plugin in self.model.objects.filter(enabled=True, plugin_type=plugin_type):
            logger.debug(f"Initializing plugin {plugin}")
            plugin.get_plugin()

    def unload_all_plugins(self):
        logger.info("Unloading all plugins")
        for plugin in PLUGIN_CACHE.plugins_list[::-1]:
            plugin._plugin_obj.remove_plugin()

        PLUGIN_CACHE.clear()

    def get_plugin(self, pk):
        plugin = self.model.objects.get(pk=pk)
        if not plugin.enabled:
            logger.warning(f"Plugin {plugin.pk} is not enabled")
            return None

        return PluginProxy(plugin.get_plugin())

    def get_plugin_by_name(self, plugin_type, name):
        plugin = self.model.objects.get(plugin_type=plugin_type, name=name)
        return self.get_plugin(plugin.pk)

    def get_plugins(self, plugin_type, plugin_name=None):
        plugins = self.model.objects.filter(plugin_type=plugin_type, enabled=True)
        if plugin_name:
            plugins = plugins.filter(plugin_name=plugin_name)

        return [self.get_plugin(p.pk) for p in plugins]


class Plugin(models.Model):  # TODO: ensure plugin unload / reload is good
    name = models.CharField(max_length=50)
    plugin_name = models.CharField(max_length=50)
    plugin_type = models.CharField(max_length=50)

    enabled = models.BooleanField(default=False)
    last_update = models.DateTimeField(auto_now=True)

    config = JSONField(blank=True, default={})

    objects = PluginManager()

    def create_plugin(self):
        PLUGIN_CREATE_LOCK.acquire()
        if self.pk in PLUGIN_CREATE_LOCKS:
            l = PLUGIN_CREATE_LOCKS[self.pk]
            PLUGIN_CREATE_LOCK.release()
            with l:
                return
        else:
            PLUGIN_CREATE_LOCKS[self.pk] = l = threading.Lock()
            l.acquire()
            PLUGIN_CREATE_LOCK.release()

        try:
            self._create_plugin()
        finally:
            with PLUGIN_CREATE_LOCK:
                del PLUGIN_CREATE_LOCKS[self.pk]
                l.release()

    def _create_plugin(self):
        logger.debug(
            f"Trying to create plugin from plugin_type:{self.plugin_type} and plugin_name:{self.plugin_name}"
        )
        plugin_class = pluginhandler.get_plugin(self.plugin_type, self.plugin_name)
        if plugin_class is None:
            logger.debug(
                f"Unknown plugin plugin_type:{self.plugin_type} and plugin_name:{self.plugin_name}"
            )
            return None

        schema = plugin_class.config_schema()
        plugin = plugin_class.__new__(plugin_class)
        plugin._related_plugins = []
        plugin.name = self.name

        config = schema.load(self.config, unknown=INCLUDE)

        def add_plugin_related(c):
            if isinstance(c, list):
                for v in c:
                    add_plugin_related(v)
            elif isinstance(c, dict):
                for v in c.values():
                    add_plugin_related(v)
            elif isinstance(c, PluginBase):
                if plugin not in c._related_plugins:
                    logger.debug(f"Adding related plugin {plugin} to {c}")
                    c._related_plugins.append(plugin)

        add_plugin_related(config)

        logger.debug(
            f"Creating plugin {self.plugin_type} / {self.plugin_name} / {self.name} with config {config} / {self.config}"
        )

        content_type = ContentType.objects.get_for_model(Plugin)
        perm, _ = Permission.objects.get_or_create(
            codename=f"{plugin.plugin_type}.{plugin.name}",
            content_type=content_type,
            defaults={
                "name": f"Can access plugin_type:{plugin.plugin_type} name:{plugin.name}"
            },
        )

        plugin.__init__(config)
        plugin._plugin_obj = self
        plugin._permission = perm

        if hasattr(plugin, "ready"):
            plugin.ready()

        PLUGIN_CACHE.add_plugin(plugin)
        plugin_loaded.send(sender=self.__class__, plugin=self)

    def is_plugin_loaded(self):
        return self in PLUGIN_CACHE

    def can_plugin_be_loaded(self):
        if self.is_plugin_loaded():
            return True

        try:
            self.get_plugin()
        except:
            logger.exception(f"Failed to load plugin {self}")
            return False
        else:
            return True

    def get_plugin(self):
        if not self.is_plugin_loaded():
            self.create_plugin()

        return PLUGIN_CACHE.get_plugin(self)

    def remove_plugin(self):
        if self.is_plugin_loaded():
            plugin = self.get_plugin()

            plugin_class = pluginhandler.get_plugin(self.plugin_type, self.plugin_name)
            schema = plugin_class.config_schema()
            config = schema.load(self.config, unknown=INCLUDE)

            def remove_plugin_related(c):
                if isinstance(c, list):
                    for v in c:
                        remove_plugin_related(v)
                elif isinstance(c, dict):
                    for v in c.values():
                        remove_plugin_related(v)
                elif isinstance(c, PluginBase):
                    if plugin in c._related_plugins:
                        c._related_plugins.remove(plugin)

            remove_plugin_related(config)

            plugin_unloaded.send(sender=self.__class__, plugin=self)
            PLUGIN_CACHE.remove_plugin(plugin)
            plugin.unload()

    def reload_plugin(self):
        logger.info(f"Reloading plugin {self}")
        self.remove_plugin()
        return self.get_plugin()

    def get_display_name(self):
        saps = self.simpleadminplugin_set.all()
        if self.config and self.config.get("display_name"):
            return self.config.get("display_name")
        elif saps and saps[0].get_display_name():
            return saps[0].get_display_name()
        else:
            return self.name

    class Meta:
        ordering = ("pk",)
        unique_together = (("name", "plugin_type"),)

    class JSONAPIMeta:
        resource_name = "plugin"

    def __str__(self):
        return f"{self.name} using {self.plugin_type}"



from collections import defaultdict


class JSONAPIObject(dict):
    """
    A JSONAPI data item.
    """

    root = None

    def __init__(self, type, id, original_object=None, links=None, populated=True):
        """
        :param type: Object type
        :param id: Object id
        :param _original_object: A dictionary that contains the original object we are trying to serialize.
                                 Used by e.g. metadata to figure out what to add.
        """
        self.id = id
        self.type = type
        self.links = links
        self._relationships = defaultdict(list)
        self._local_relationships = defaultdict(list)
        self._original_object = original_object
        self._populated = populated

    def serialize(self, request):
        """
        Turns this object into a plain JSON API object ready to be JSON encoded.
        """
        obj = {"type": self.type, "id": self.id}
        if self._populated:
            obj["attributes"] = dict(self)

        relationships = {}
        for relationship_type, relationship_list in self._relationships.items():
            r = []
            for relationship in relationship_list:
                r.append({"type": relationship.type, "id": relationship.id})
            if r:
                relationships[relationship_type] = {"data": r}

        for relationship_type, relationship_list in self._local_relationships.items():
            r = []
            for relationship in relationship_list:
                r.append(relationship.serialize(request))

            if r:
                relationships[relationship_type] = {"data": r}

        if relationships:
            obj["relationships"] = relationships

        links = {}

        if self._original_object and hasattr(self._original_object, "get_absolute_url"):
            url = self._original_object.get_absolute_url()
            if url:
                url = request.build_absolute_uri(url)
                if url:
                    links["self"] = url

        if isinstance(links.get("self"), str) and links["self"].startswith("/"):
            links["self"] = request.build_absolute_uri(links["self"])

        if self._original_object and hasattr(
            self._original_object, "get_additional_urls"
        ):
            urls = self._original_object.get_additional_urls()
            for k, url in urls.items():
                url = request.build_absolute_uri(url)
                if url:
                    links[k] = url

        if self.links:
            links.update(self.links)

        if links:
            obj["links"] = links

        return obj

    def add_relationship(self, type, obj, local=False):
        if local:
            self._local_relationships[type].append(obj)
        else:
            self._relationships[type].append(obj)
            self.root.add_included(obj)


class JSONAPIRoot:
    """
    A base to serialize listings to proper JSONAPI format.
    """

    def __init__(self):
        self.data = []
        self.included = {}  # ('type', 'id') keys mapped to a JSONAPIObject
        self.links = {}
        self.meta = {}
        self.errors = {}

    def append(self, obj):
        """
        Appends a new object to the output
        """
        obj.root = self
        self.data.append(obj)

    def add_included(self, obj):
        """
        Appends a new object that should be included (and referenced by objects in data)
        """
        obj.root = self
        self.included[(obj.type, obj.id)] = obj

    def _serialize_success(self, request):
        data = []
        for obj in self.data:
            data.append(obj.serialize(request))

        included = []
        for obj in self.included.values():
            included.append(obj.serialize(request))

        if len(data) == 1:
            data = data[0]

        obj = {"data": data, "included": included}

        if self.meta:
            obj["meta"] = self.meta

        if self.links:
            obj["links"] = self.links

        return obj

    def _serialize_errors(self, request):
        errors = self.errors.copy()

        if self.meta:
            errors["meta"] = self.meta

        if self.links:
            errors["links"] = self.links

        obj = {"errors": [errors]}

        return obj

    def serialize(self, request):
        """
        Turns the root into a JSONAPI object ready to be JSON encoded.
        """
        if self.errors:
            return self._serialize_errors(request)
        else:
            return self._serialize_success(request)

    @classmethod
    def success_status(cls, message_or_meta):
        obj = cls()
        if isinstance(message_or_meta, dict):
            obj.meta.update(message_or_meta)
        else:
            obj.meta["message"] = message_or_meta
        return obj

    @classmethod
    def error_status(
        cls,
        id_=None,
        status=None,
        code=None,
        title=None,
        detail=None,
        source_param=None,
    ):
        obj = cls()

        if id_:
            obj.errors["id"] = id_

        if status:
            obj.errors["status"] = status

        if code:
            obj.errors["code"] = code

        if title:
            obj.errors["title"] = title

        if detail:
            obj.errors["detail"] = detail

        if source_param:
            obj.errors["source"] = {"parameter": source_param}

        return obj



from .log import Log, LogMessage, LogNotificationComponent
from .plugin import Plugin, PluginCache
from .scheduler import FailedToParseScheduleException, Schedule, parse_schedule_trigger

__all__ = [
    "Log",
    "LogNotificationComponent",
    "LogMessage",
    "PluginCache",
    "Plugin",
    "Schedule",
    "FailedToParseScheduleException",
    "parse_schedule_trigger",
]



import logging

from django.contrib.auth.models import Permission, User
from django.contrib.auth.password_validation import validate_password
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from rest_framework import permissions, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from ....models import Plugin
from .shared import ADMIN_RENDERER_CLASSES, ServiceAwareHyperlinkedIdentityField

logger = logging.getLogger(__name__)


class UserPasswordSerializer(serializers.Serializer):
    password = serializers.CharField()


class UserPermissionsSerializer(serializers.Serializer):
    permissions = serializers.ListField(child=serializers.CharField())


class UserSerializer(serializers.HyperlinkedModelSerializer):
    url = ServiceAwareHyperlinkedIdentityField(view_name="user-detail")
    date_joined = serializers.DateTimeField(read_only=True)
    last_login = serializers.DateTimeField(read_only=True)
    permissions = serializers.SerializerMethodField()

    class Meta:
        model = User
        exclude = ("password", "groups", "user_permissions")

    class JSONAPIMeta:
        resource_name = "user"

    def get_permissions(self, obj):
        content_type = ContentType.objects.get_for_model(Plugin)
        return list(
            obj.user_permissions.filter(content_type=content_type).values_list(
                "codename", flat=True
            )
        )


class UserModelView(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer
    renderer_classes = ADMIN_RENDERER_CLASSES
    permission_classes = (permissions.IsAdminUser,)
    pagination_class = None

    service = None

    @action(methods=["post"], detail=True)
    def set_password(self, request, pk=None):
        user = self.get_object()
        serializer = UserPasswordSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        password = serializer.data["password"]
        try:
            validate_password(password, user)
        except ValidationError:
            return Response(
                {"message": "Please choose a more complex password."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        user.set_password(password)
        user.save()

        return Response({"message": "User password changed successfully"})

    @action(methods=["post"], detail=True)
    def set_permissions(self, request, pk=None):
        user = self.get_object()
        serializer = UserPermissionsSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        content_type = ContentType.objects.get_for_model(Plugin)
        current_permissions = set(
            user.user_permissions.filter(content_type=content_type).values_list(
                "codename", flat=True
            )
        )
        permissions = set(serializer.data["permissions"])

        for codename in current_permissions - permissions:
            try:
                permission = Permission.objects.get(
                    content_type=content_type, codename=codename
                )
            except:
                logger.warning("Unknown permission: %s" % (codename))
                continue

            user.user_permissions.remove(permission)

        for codename in permissions - current_permissions:
            try:
                permission = Permission.objects.get(
                    content_type=content_type, codename=codename
                )
            except:
                logger.warning("Unknown permission: %s" % (codename))
                continue

            user.user_permissions.add(permission)

        return Response(
            {"status": "success", "message": "Permissions changed successfully"}
        )



import logging

from rest_framework import permissions
from rest_framework.response import Response
from rest_framework.views import APIView

from ...jsonapi import JSONAPIObject, JSONAPIRoot
from ...models import Plugin

logger = logging.getLogger(__name__)


class APIConfigView(APIView):
    service = None

    permission_classes = (permissions.AllowAny,)

    def get(self, request):
        root = JSONAPIRoot()

        for plugin in Plugin.objects.filter(enabled=True, plugin_type="service"):
            links = {"self": request.build_absolute_uri("/%s/" % plugin.name)}
            plugin_type = "%s_%s" % (plugin.plugin_type, plugin.plugin_name)

            plugin_obj = plugin.get_plugin()
            permission_obj = JSONAPIObject("permission", "access_%s" % (plugin.name,))
            urls = plugin_obj.get_urls()

            permission_obj["can_access"] = False
            if urls:
                for pattern in urls:
                    view_pattern = pattern.resolve("")
                    if view_pattern:
                        func = view_pattern.func
                        view = func.view_class(**func.view_initkwargs)
                        permissions = [
                            permission() for permission in view.permission_classes
                        ]
                        for permission in permissions:
                            if permission.has_permission(request, view):
                                permission_obj["can_access"] = True
                        break
            else:  # TODO: add websocket support
                permission_obj["can_access"] = True

            if not permission_obj[
                "can_access"
            ]:  # Not sure if we should leak inaccessible services...
                continue

            obj = JSONAPIObject(plugin_type, plugin.name, links=links)
            obj["display_name"] = plugin.config.get("display_name", plugin.name)
            root.append(obj)

            obj.add_relationship("permission", permission_obj, local=True)

        return Response(root.serialize(request))



from marshmallow import fields


def handle_length(schema, field, validator, parent_schema):
    """Adds validation logic for ``marshmallow.validate.Length``, setting the
    values appropriately for ``fields.List``, ``fields.Nested``, and
    ``fields.String``.

    Args:
        schema (dict): The original JSON schema we generated. This is what we
            want to post-process.
        field (fields.Field): The field that generated the original schema and
            who this post-processor belongs to.
        validator (marshmallow.validate.Length): The validator attached to the
            passed in field.
        parent_schema (marshmallow.Schema): The Schema instance that the field
            belongs to.

    Returns:
        dict: A, possibly, new JSON Schema that has been post processed and
            altered.

    Raises:
        ValueError: Raised if the `field` is something other than
            `fields.List`, `fields.Nested`, or `fields.String`
    """
    if isinstance(field, fields.String):
        minKey = "minLength"
        maxKey = "maxLength"
    elif isinstance(field, (fields.List, fields.Nested)):
        minKey = "minItems"
        maxKey = "maxItems"
    else:
        raise ValueError(
            "In order to set the Length validator for JSON "
            "schema, the field must be either a List or a String"
        )

    if validator.min:
        schema[minKey] = validator.min

    if validator.max:
        schema[maxKey] = validator.max

    if validator.equal:
        schema[minKey] = validator.equal
        schema[maxKey] = validator.equal

    return schema


def handle_one_of(schema, field, validator, parent_schema):
    """Adds the validation logic for ``marshmallow.validate.OneOf`` by setting
    the JSONSchema `enum` property to the allowed choices in the validator.

    Args:
        schema (dict): The original JSON schema we generated. This is what we
            want to post-process.
        field (fields.Field): The field that generated the original schema and
            who this post-processor belongs to.
        validator (marshmallow.validate.OneOf): The validator attached to the
            passed in field.
        parent_schema (marshmallow.Schema): The Schema instance that the field
            belongs to.

    Returns:
        dict: A, possibly, new JSON Schema that has been post processed and
            altered.
    """
    if validator.choices:
        schema["enum"] = list(validator.choices)
        schema["enumNames"] = list(validator.labels)

    return schema


def handle_range(schema, field, validator, parent_schema):
    """Adds validation logic for ``marshmallow.validate.Range``, setting the
    values appropriately ``fields.Number`` and it's subclasses.

    Args:
        schema (dict): The original JSON schema we generated. This is what we
            want to post-process.
        field (fields.Field): The field that generated the original schema and
            who this post-processor belongs to.
        validator (marshmallow.validate.Length): The validator attached to the
            passed in field.
        parent_schema (marshmallow.Schema): The Schema instance that the field
            belongs to.

    Returns:
        dict: A, possibly, new JSON Schema that has been post processed and
            altered.
    """
    if not isinstance(field, fields.Number):
        return schema

    if validator.min:
        schema["minimum"] = validator.min

    if validator.max:
        schema["maximum"] = validator.max

    return schema



from marshmallow import Schema as OriginalSchema
from marshmallow import INCLUDE, SchemaOpts, ValidationError, fields

__all__ = ["INCLUDE", "Schema", "fields", "ValidationError"]


class UISchemaOpts(SchemaOpts):
    def __init__(self, meta, **kwargs):
        SchemaOpts.__init__(self, meta, **kwargs)
        self.ui_schema = getattr(meta, "ui_schema", {})


class Schema(OriginalSchema):
    OPTIONS_CLASS = UISchemaOpts


Schema.Meta.unknown = INCLUDE



from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from rest_framework import permissions, serializers, viewsets

from ....models import Plugin
from .shared import ADMIN_RENDERER_CLASSES, ServiceAwareHyperlinkedIdentityField


class PermissionSerializer(serializers.HyperlinkedModelSerializer):
    url = ServiceAwareHyperlinkedIdentityField(view_name="permission-detail")

    class Meta:
        model = Permission
        fields = ("name", "codename", "url")

    class JSONAPIMeta:
        resource_name = "permission"


class PermissionModelView(viewsets.ReadOnlyModelViewSet):
    serializer_class = PermissionSerializer
    renderer_classes = ADMIN_RENDERER_CLASSES
    permission_classes = (permissions.IsAdminUser,)
    pagination_class = None

    service = None

    def get_queryset(self):
        return Permission.objects.filter(
            content_type=ContentType.objects.get_for_model(Plugin),
            codename__startswith="service.",
        )



# Generated by Django 2.2 on 2019-05-18 15:57

import django.db.models.deletion
import jsonfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("unplugged", "0004_auto_20190425_1302")]

    operations = [
        migrations.AlterField(
            model_name="simpleadminplugin",
            name="auto_created_plugins",
            field=models.ManyToManyField(
                blank=True, related_name="auto_created_plugins", to="unplugged.Plugin"
            ),
        ),
        migrations.AlterField(
            model_name="simpleadminplugin",
            name="config",
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),
        migrations.CreateModel(
            name="Schedule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("interval", models.IntegerField()),
                ("command", models.CharField(max_length=200)),
                ("kwargs", jsonfield.fields.JSONField()),
                ("enabled", models.BooleanField(default=False)),
                (
                    "plugin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="unplugged.Plugin",
                    ),
                ),
            ],
        ),
    ]



from django.conf import settings

from ..baseplugin import PluginBase


def pytest_configure():
    settings.configure(
        DATABASES={
            "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}
        },
        INSTALLED_APPS=[
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "unplugged",
        ],
    )

    class VehiclePlugin(PluginBase):
        pass



from .log import LogModelView
from .permission import PermissionModelView
from .plugin import PluginBaseListView, PluginModelView
from .scheduler import ScheduleModelView
from .simpleadmin import (
    ShowAdminUrlsView,
    SimpleAdminPluginModelView,
    SimpleAdminTemplateModelView,
)
from .user import UserModelView

__all__ = [
    "LogModelView",
    "PermissionModelView",
    "PluginBaseListView",
    "PluginModelView",
    "ScheduleModelView",
    "SimpleAdminTemplateModelView",
    "SimpleAdminPluginModelView",
    "ShowAdminUrlsView",
    "UserModelView",
]



# Generated by Django 2.2 on 2019-07-06 08:17

import django.db.models.deletion
import jsonfield.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("unplugged", "0006_auto_20190522_1907"),
    ]

    operations = [
        migrations.CreateModel(
            name="Log",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("action", models.CharField(db_index=True, max_length=50)),
                ("start_datetime", models.DateTimeField(auto_now_add=True)),
                ("end_datetime", models.DateTimeField(null=True)),
                ("progress", models.PositiveIntegerField(default=0)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("pending", "Pending"),
                            ("ongoing", "Ongoing"),
                            ("success", "Success"),
                            ("failed", "Failed"),
                            ("cancelled", "Cancelled"),
                        ],
                        default="pending",
                        max_length=12,
                    ),
                ),
                (
                    "plugin",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="unplugged.Plugin",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AlterField(
            model_name="schedule",
            name="kwargs",
            field=jsonfield.fields.JSONField(blank=True),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="plugin_unique_id",
            field=models.CharField(
                blank=True, db_index=True, max_length=100, null=True
            ),
        ),
        migrations.CreateModel(
            name="LogMessage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime", models.DateTimeField(auto_now_add=True)),
                ("msg", models.TextField(blank=True, default="")),
                (
                    "log",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="unplugged.Log"
                    ),
                ),
            ],
        ),
    ]



from django.conf.urls import url

from ...schema import Schema
from ...plugins import ServicePlugin
from .views import APIConfigView


class APIServicePlugin(ServicePlugin):
    plugin_name = "api"
    config_schema = Schema

    def get_urls(self):
        return [url("^/?$", APIConfigView.as_view(service=self))]

    def unload(self):
        pass



# Generated by Django 2.2 on 2019-05-22 19:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("unplugged", "0005_auto_20190518_1557")]

    operations = [
        migrations.RemoveField(model_name="schedule", name="interval"),
        migrations.AddField(
            model_name="schedule",
            name="method",
            field=models.CharField(
                choices=[("cron", "Cron"), ("interval", "Interval")],
                default="cron",
                max_length=10,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="schedule",
            name="method_config",
            field=models.CharField(default="", max_length=500),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="schedule",
            name="plugin_unique_id",
            field=models.CharField(db_index=True, max_length=100, null=True),
        ),
    ]



import logging
import re

logger = logging.getLogger(__name__)


def register_channels(channels):
    logger.debug("Registering channels: %r" % (channels,))
    urlpatterns.extend(channels)


def unregister_channel(pattern):
    logger.debug("Unregistering pattern: %s" % (pattern,))
    pattern = re.compile(pattern)
    for i, p in enumerate(urlpatterns):
        if pattern == p.pattern.regex:
            break
    else:
        return

    urlpatterns.pop(i)


def clear_channels():
    logger.debug("Clearing channels")
    urlpatterns[:] = []


urlpatterns = []



from .schema import fields


def dump_ui_schema(obj):
    ui_schema = {}
    ui_schema.update(getattr(obj.opts, "ui_schema", {}))
    for field_name, field in sorted(obj.fields.items()):
        if isinstance(field, fields.Nested):
            nested_ui_schema = dump_ui_schema(field.nested())
            if field.many:
                nested_ui_schema = {"items": nested_ui_schema}
            field_ui_schema = nested_ui_schema
        else:
            field_ui_schema = field.metadata.get("ui_schema")

        if field_ui_schema:
            ui_schema[field_name] = field_ui_schema

    return ui_schema



import inspect
import logging
from functools import wraps

from django.http import HttpResponse
from rest_framework import serializers, status
from rest_framework.response import Response

from .jsonapi import JSONAPIObject, JSONAPIRoot
from .jsonschema import dump_ui_schema
from .schema import Schema, ValidationError, INCLUDE
from .libs.marshmallow_jsonschema import JSONSchema

logger = logging.getLogger(__name__)


class CommandViewMixin:
    def call_command(self, request, obj, additional_kwargs=None):
        additional_kwargs = additional_kwargs or {}
        serializer = CommandSerializer(data=request.data)
        if not serializer.is_valid():
            jsonapi_root = JSONAPIRoot.error_status(
                id_="deserialize_failed", detail="Failed to deserialize your request"
            )
            jsonapi_root.meta.update(serializer.errors)
            return Response(
                jsonapi_root.serialize(request), status=status.HTTP_400_BAD_REQUEST
            )

        command_name = serializer.validated_data["name"]
        command = obj.get_command(command_name)
        if not command:
            jsonapi_root = JSONAPIRoot.error_status(
                id_="unknown_command", detail=f"{command_name} is not a known command"
            )
            jsonapi_root.meta.update(serializer.errors)
            return Response(
                jsonapi_root.serialize(request), status=status.HTTP_400_BAD_REQUEST
            )

        kwargs = serializer.validated_data["kwargs"]
        try:
            kwargs = command.parse_kwargs(kwargs)
        except ValidationError as err:
            jsonapi_root = JSONAPIRoot.error_status(
                id_="invalid_args",
                detail=f"You provided invalid arguments to the function {command}",
            )
            jsonapi_root.meta.update(err.messages)
            return Response(
                jsonapi_root.serialize(request), status=status.HTTP_400_BAD_REQUEST
            )

        kwargs["self"] = obj
        kwargs.update(additional_kwargs)
        if command.need_request:
            kwargs["request"] = request

        try:
            command_result = command.execute(kwargs)
        except Exception:
            logger.exception(f"Failed to execute {command} with args {kwargs}")
            jsonapi_root = JSONAPIRoot.error_status(
                id_="execution_failed", detail="Failed to execute command"
            )
            return Response(
                jsonapi_root.serialize(request), status=status.HTTP_400_BAD_REQUEST
            )

        if isinstance(command_result, HttpResponse):
            return command_result

        if isinstance(command_result, JSONAPIRoot):
            return Response(command_result.serialize(request))

        jsonapi_root = JSONAPIRoot.success_status(command_result)
        return Response(jsonapi_root.serialize(request))


class CommandBase:
    @classmethod
    def get_jsonapi_commands(cls):
        commands = getattr(cls, "__commands__", None) or {}
        serialized_commands = []
        for command_obj, command in zip(
            CommandSerializer(commands.values(), many=True).data, commands.values()
        ):
            obj = JSONAPIObject("command", command.name, command)
            obj.update(command_obj)
            serialized_commands.append(obj)
        return serialized_commands

    def get_command(self, command):
        return self.__commands__.get(command)


class CommandBaseMeta(type):
    def __new__(mcls, name, bases, namespace):
        if CommandBase not in bases:
            bases += (CommandBase,)
        cls = super(CommandBaseMeta, mcls).__new__(mcls, name, bases, namespace)
        cls.__commands__ = {}
        for fn_name, fn in inspect.getmembers(cls, predicate=inspect.isfunction):
            command = getattr(fn, "__command__", None)
            if command:
                cls.__commands__[command.name] = command

        return cls


class Command:
    def __init__(
        self,
        fn,
        name=None,
        display_name=None,
        description=None,
        schema=None,
        metadata=None,
        need_request=False,
    ):
        self.fn = fn
        if name:
            self.name = name
        else:
            if fn.__name__.startswith("command_"):
                self.name = fn.__name__[8:]
            else:
                self.name = fn.__name__

        self.display_name = display_name or self.name
        self.description = description or "No description found"
        self.schema = schema or Schema
        self.metadata = metadata or {}
        self.need_request = need_request

    def parse_kwargs(self, kwargs):
        return self.schema().load(kwargs, unknown=INCLUDE)

    def execute(self, kwargs):
        try:
            result = self.fn(**kwargs)
        except:
            raise
        else:
            return result


def command(
    name=None,
    display_name=None,
    description=None,
    schema=None,
    metadata=None,
    need_request=False,
):
    def decorator(fn):
        fn.__command__ = Command(
            fn, name, display_name, description, schema, metadata, need_request
        )

        @wraps(fn)
        def decorated(*args, **kwargs):
            return fn(*args, **kwargs)

        return decorated

    return decorator


class CommandSerializer(serializers.Serializer):
    command = serializers.CharField(source="name")
    kwargs = serializers.DictField(required=False, write_only=True)
    display_name = serializers.CharField(read_only=True)
    description = serializers.CharField(read_only=True)
    schema = serializers.SerializerMethodField(read_only=True)
    ui_schema = serializers.SerializerMethodField(read_only=True)

    def get_schema(self, obj):
        schema = obj.schema()
        json_schema = JSONSchema()
        return json_schema.dump(schema)

    def get_ui_schema(self, obj):
        schema = obj.schema()
        return dump_ui_schema(schema)



import sys

import marshmallow

PY2 = int(sys.version_info[0]) == 2

if PY2:
    text_type = unicode
    binary_type = str
    basestring = basestring
else:
    text_type = str
    binary_type = bytes
    basestring = (str, bytes)


if marshmallow.__version__.split(".", 1)[0] >= "3":
    marshmallow_2 = False

    def dot_data_backwards_compatible(json_schema):
        return json_schema


else:
    marshmallow_2 = True

    def dot_data_backwards_compatible(json_schema):
        return json_schema.data


__all__ = (
    "text_type",
    "binary_type",
    "basestring",
    "marshmallow_2",
    "dot_data_backwards_compatible",
)



from django.contrib import admin

from .models import Log, LogMessage, Plugin, Schedule

admin.site.register(Plugin)
admin.site.register(Schedule)
admin.site.register(Log)
admin.site.register(LogMessage)



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "unplugged.services.user"
    verbose_name = "User Service"
    label = "services_user"

    def ready(self):
        from .handler import UserServicePlugin  # NOQA



import json
import logging

from django.conf import settings
from django.db import models
from django.db.models.signals import post_save
from django.utils.timezone import now

from autobahn.twisted.wamp import ApplicationSession

from .plugin import Plugin

logger = logging.getLogger(__name__)


class LogManager(models.Manager):
    def start_chain(self, plugin, action, user=None):
        if plugin is not None:
            if not isinstance(plugin, Plugin):
                plugin = plugin._plugin_obj
        logger.debug(
            f"New log chain started for plugin:{plugin!r} user:{user!r} action:{action}"
        )
        log = self.create(user=user, action=action, plugin=plugin)
        return LogChain(log)

    def cleanup_hanging_chains(self):
        logger.debug("Clearing up hanging log chains")
        self.filter(end_datetime__isnull=True).update(
            end_datetime=now(), status=Log.STATUS_FAILED
        )


class Log(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL
    )
    plugin = models.ForeignKey(Plugin, null=True, on_delete=models.SET_NULL)
    action = models.CharField(max_length=50, db_index=True)
    start_datetime = models.DateTimeField(auto_now_add=True)
    end_datetime = models.DateTimeField(null=True)
    progress = models.PositiveIntegerField(default=0)

    STATUS_PENDING = "pending"
    STATUS_ONGOING = "ongoing"
    STATUS_SUCCESS = "success"
    STATUS_FAILED = "failed"
    STATUS_CANCELLED = "cancelled"

    status = models.CharField(
        default=STATUS_PENDING,
        choices=(
            (STATUS_PENDING, "Pending"),
            (STATUS_ONGOING, "Ongoing"),
            (STATUS_SUCCESS, "Success"),
            (STATUS_FAILED, "Failed"),
            (STATUS_CANCELLED, "Cancelled"),
        ),
        max_length=12,
    )

    objects = LogManager()

    def __str__(self):
        return f"user:{self.user!r} plugin:{self.plugin!r} progress:{self.progress} status:{self.status}"

    class JSONAPIMeta:
        resource_name = "log"


class LogMessage(models.Model):
    log = models.ForeignKey(
        Log, db_index=True, on_delete=models.CASCADE, related_name="log_messages"
    )
    datetime = models.DateTimeField(auto_now_add=True)
    msg = models.TextField(blank=True, default="")

    class Meta:
        ordering = ("-datetime",)

    class JSONAPIMeta:
        resource_name = "logmessage"


class LogChain:
    def __init__(self, log):
        self._log = log

    def log(self, progress=None, msg=""):
        modified = False

        if not isinstance(msg, str):
            msg = json.dumps(msg)

        if self._log.status == self._log.STATUS_PENDING:
            self._log.status = self._log.STATUS_ONGOING
            modified = True

        if progress is not None and self._log.progress != progress:
            self._log.progress = progress
            modified = True

        if modified:
            self._log.save()

        if msg:
            LogMessage.objects.create(log=self._log, msg=msg)

    def finish_chain(self, status):
        self._log.status = status
        self._log.end_datetime = now()
        self._log.save()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            status = Log.STATUS_FAILED
            self.log(msg=str(traceback))
        else:
            status = Log.STATUS_SUCCESS

        self.finish_chain(status)


class LogNotificationComponent(ApplicationSession):
    def __init__(self, config=None):
        ApplicationSession.__init__(self, config)
        post_save.connect(self.new_log, sender=Log)
        post_save.connect(self.new_log_message, sender=LogMessage)

    def new_log(self, sender, instance, created, raw, *args, **kwargs):
        from ..views.log import LogSerializer

        self.publish(settings.WAMP_LOG_TOPIC, LogSerializer(instance).data)

    def new_log_message(self, sender, instance, created, raw, *args, **kwargs):
        from ..views.log import LogMessageSerializer

        self.publish(
            f"{settings.WAMP_LOG_TOPIC}.{instance.log_id}",
            LogMessageSerializer(instance).data,
        )



# Generated by Django 2.2 on 2019-04-03 10:26

import jsonfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Plugin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                ("plugin_name", models.CharField(max_length=50)),
                ("plugin_type", models.CharField(max_length=50)),
                ("enabled", models.BooleanField(default=False)),
                ("last_update", models.DateTimeField(auto_now=True)),
                ("config", jsonfield.fields.JSONField(blank=True, default={})),
            ],
            options={"ordering": ("pk",), "unique_together": {("name", "plugin_type")}},
        )
    ]



import logging

from rest_framework import permissions, serializers, viewsets
from rest_framework.decorators import action
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework.response import Response
from rest_framework_json_api.renderers import JSONRenderer

from ..commands import CommandViewMixin
from ..jsonschema import dump_ui_schema
from ..libs.marshmallow_jsonschema import JSONSchema
from ..models import Plugin
from ..pluginhandler import pluginhandler

ADMIN_RENDERER_CLASSES = (JSONRenderer, BrowsableAPIRenderer)

logger = logging.getLogger(__name__)


class PluginBaseSerializer(serializers.Serializer):
    id = serializers.SerializerMethodField()
    plugin_name = serializers.CharField()
    plugin_type = serializers.CharField()
    schema = serializers.SerializerMethodField()
    ui_schema = serializers.SerializerMethodField()
    commands = serializers.SerializerMethodField()
    traits = serializers.SerializerMethodField()
    default_permission = serializers.CharField(default=None)

    class JSONAPIMeta:
        resource_name = "plugin_base"

    def get_commands(self, obj):
        if hasattr(obj, "get_jsonapi_commands"):
            return obj.get_jsonapi_commands()
        else:
            return []

    def get_schema(self, obj):
        schema = obj.config_schema()
        json_schema = JSONSchema()

        return json_schema.dump(schema)

    def get_ui_schema(self, obj):
        schema = obj.config_schema()
        return dump_ui_schema(schema)

    def get_traits(self, obj):
        if hasattr(obj, "__trait__"):
            return obj.__trait__
        else:
            return []

    def id(self, obj):
        return f"{obj.plugin_type}:{obj.plugin_name}"


class PluginBaseListView(viewsets.ViewSet):
    renderer_classes = ADMIN_RENDERER_CLASSES
    serializer_class = PluginBaseSerializer
    permission_classes = (permissions.IsAdminUser,)
    pagination_class = None

    resource_name = "plugin_base"

    def list(self, request):
        pluginbases = list(pluginhandler.get_all_plugins())
        serializer = self.serializer_class(pluginbases, many=True)
        return Response(serializer.data)


class PluginSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="plugin-detail")
    config = serializers.JSONField()

    class Meta:
        model = Plugin
        fields = "__all__"

    class JSONAPIMeta:
        resource_name = "plugin"


class PluginModelView(viewsets.ModelViewSet, CommandViewMixin):
    queryset = Plugin.objects.all()
    serializer_class = PluginSerializer
    renderer_classes = ADMIN_RENDERER_CLASSES
    permission_classes = (permissions.IsAdminUser,)
    pagination_class = None

    @action(methods=["post"], detail=False, url_path="reload", url_name="reload")
    def reload_all(self, request):
        Plugin.objects.unload_all_plugins()
        Plugin.objects.bootstrap()

        return Response({"status": "success", "message": "Modules reloaded"})

    @action(methods=["post"], detail=True)
    def command(self, request, pk=None):
        plugin = self.get_object().get_plugin()

        return self.call_command(request, plugin)

    @action(methods=["post"], detail=True, url_path="reload", url_name="reload-plugin")
    def reload_plugin(self, request, pk=None):
        plugin = self.get_object()
        plugin.reload_plugin()
        return Response({"message": "Plugin reloaded"})



import logging

from django.conf import settings
from django.db.models.signals import post_delete, post_save

from .models import Schedule
from .signals import plugin_loaded, plugin_unloaded

logger = logging.getLogger(__name__)


class ScheduleManager:
    def __init__(self):
        self.schedules = {}

    def plugin_loaded(self, sender, plugin, **kwargs):
        logger.debug(f"Creating schedules for {plugin}")
        for schedule in Schedule.objects.filter(plugin=plugin, enabled=True):
            self.reload_schedule(schedule)

    def plugin_unloaded(self, sender, plugin, **kwargs):
        logger.debug(f"Removing schedules for {plugin}")
        for schedule in Schedule.objects.filter(plugin=plugin, enabled=True):
            self.unload_schedule(schedule)

    def unload_schedule(self, schedule):
        logger.debug(f"Unloading schedule {schedule}")
        schedules = self.schedules.setdefault(schedule.plugin.pk, {})
        job = schedules.pop(schedule.pk, None)
        if job:
            job.remove()

    def schedule_modified(self, sender, instance, created, **kwargs):
        self.reload_schedule(instance)

    def schedule_deleted(self, sender, instance, **kwargs):
        self.unload_schedule(instance)

    def load_schedule(self, schedule):
        logger.debug(f"Loading schedule {schedule}")

        job = settings.SCHEDULER.add_job(
            self.trigger_schedule,
            schedule.get_trigger(),
            id=f"scheduler_{schedule.pk}",
            kwargs={"pk": schedule.pk},
        )
        schedules = self.schedules.setdefault(schedule.plugin.pk, {})
        schedules[schedule.pk] = job

    def reload_schedule(self, schedule):
        self.unload_schedule(schedule)
        if not schedule.enabled:
            return

        self.load_schedule(schedule)

    def trigger_schedule(self, pk):
        schedule = Schedule.objects.get(pk=pk)
        if not schedule.enabled:
            logger.warning(f"Trying to trigger disabled schedule {schedule.pk}")
            return

        plugin = schedule.plugin.get_plugin()
        command = plugin.get_command(schedule.command)
        kwargs = schedule.kwargs or {}
        kwargs = command.parse_kwargs(kwargs)
        kwargs["self"] = plugin

        try:
            _ = command.execute(kwargs)
        except Exception:
            logger.exception(
                f"Failed to execute {command} with args {kwargs} from schedule"
            )

    def start(self):
        logger.debug("Started schedule manager")
        post_delete.connect(self.schedule_deleted, sender=Schedule)
        post_save.connect(self.schedule_modified, sender=Schedule)
        plugin_loaded.connect(self.plugin_loaded)
        plugin_unloaded.connect(self.plugin_unloaded)

    def stop(self):
        logger.debug("Stopped schedule manager")
        post_delete.disconnect(self.schedule_deleted, sender=Schedule)
        post_save.disconnect(self.schedule_modified, sender=Schedule)
        plugin_loaded.disconnect(self.plugin_loaded)
        plugin_unloaded.disconnect(self.plugin_unloaded)


schedule_manager = ScheduleManager()



# Generated by Django 2.0.4 on 2019-07-15 19:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("unplugged", "0007_auto_20190706_0817")]

    operations = [
        migrations.RemoveField(
            model_name="simpleadminplugin", name="auto_created_plugins"
        ),
        migrations.RemoveField(model_name="simpleadminplugin", name="plugin"),
        migrations.RemoveField(model_name="simpleadminplugin", name="template"),
        migrations.DeleteModel(name="SimpleAdminTemplate"),
        migrations.AlterModelOptions(
            name="logmessage", options={"ordering": ("-datetime",)}
        ),
        migrations.AlterField(
            model_name="logmessage",
            name="log",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="log_messages",
                to="unplugged.Log",
            ),
        ),
        migrations.DeleteModel(name="SimpleAdminPlugin"),
    ]



import logging
from abc import ABCMeta, abstractproperty

from .commands import CommandBaseMeta
from .schema import fields
from .pluginhandler import pluginhandler

logger = logging.getLogger(__name__)


def register_plugin(cls):
    """Autoregisters the plugin with the handler"""
    plugin_name = getattr(cls, "plugin_name", None)
    plugin_type = getattr(cls, "plugin_type", None)

    if isinstance(plugin_type, str):
        if isinstance(plugin_name, str):
            pluginhandler.register_plugin(cls)
        else:
            pluginhandler.register_plugin_type(cls)


class PluginBaseMeta(ABCMeta, CommandBaseMeta):
    def __new__(mcls, name, bases, namespace):
        cls = super(PluginBaseMeta, mcls).__new__(mcls, name, bases, namespace)
        register_plugin(cls)
        return cls

    @property
    def pk(self):
        return "%s:%s" % (self.plugin_type, self.plugin_name)


class PluginBase(metaclass=PluginBaseMeta):
    __commands__ = None
    __traits__ = []

    _related_plugins = None
    _plugin_obj = None

    name = None

    def __init__(self, config):
        """
        Instantiates the plugin with a config
        """
        self.config = config

    @abstractproperty
    def plugin_type(self):
        """
        Name of plugin type
        """
        raise NotImplementedError

    @abstractproperty
    def plugin_name(self):
        """
        Name of plugin
        """
        raise NotImplementedError

    def unload(self):
        """
        Unloads the plugin instance and shuts it down.
        """
        pass

    @abstractproperty
    def config_schema(self):
        """
        Marshmallow Schema used to configure the plugin.
        """
        raise NotImplementedError


class RelatedPluginField(fields.Field):
    def _jsonschema_type_mapping(self):
        from .models import Plugin

        plugins = Plugin.objects.filter(enabled=True)

        plugin_type = getattr(self.metadata.get("plugin_type"), "plugin_type", None)
        if plugin_type:
            plugins = plugins.filter(plugin_type=plugin_type)

        plugin_name = getattr(self.metadata.get("plugin_name"), "plugin_name", None)
        if plugin_name:
            plugins = plugins.filter(plugin_name=plugin_name)

        traits_required = set(self.metadata.get("traits", []))
        plugins = [
            p
            for p in plugins
            if p.can_plugin_be_loaded()
            and not traits_required - set(p.get_plugin().__traits__)
        ]

        if not plugins:
            return {"type": "number", "enum": [0], "enumNames": ["No plugins added"]}

        return {
            "type": "number",
            "enum": [p.pk for p in plugins],
            "enumNames": [p.get_display_name() for p in plugins],
        }

    def _deserialize(self, value, attr, data, **kwargs):
        from .models import Plugin

        logger.debug("Trying to resolve related %r" % (value,))
        try:
            return Plugin.objects.get_plugin(pk=value)
        except Plugin.DoesNotExist:
            return None


class DjangoModelField(fields.Field):
    def _jsonschema_type_mapping(self):
        queryset = self.metadata.get("queryset", [])
        name_field = self.metadata.get("name_field", "pk")
        objs = [obj for obj in queryset]

        return {
            "type": "number",
            "enum": [obj.pk for obj in objs],
            "enumNames": [getattr(obj, name_field) for obj in objs],
        }

    def _deserialize(self, value, attr, data, **kwargs):
        queryset = self.metadata.get("queryset", [])
        logger.debug("Trying to resolve object %r" % (value,))
        try:
            return queryset.get(pk=value)
        except queryset.model.DoesNotExist:
            return None



from rest_framework import serializers
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework_json_api.renderers import JSONRenderer

__all__ = ["ServiceAwareHyperlinkedIdentityField", "ADMIN_RENDERER_CLASSES"]

ADMIN_RENDERER_CLASSES = (JSONRenderer, BrowsableAPIRenderer)


class ServiceAwareHyperlinkedIdentityField(serializers.HyperlinkedIdentityField):
    _view_name = None

    @property
    def view_name(self):
        if hasattr(self, "parent"):
            module = self.context["view"].__class__.__module__.split(".")[0]
            return f"{module}:{self.context['view'].service.name}:{self._view_name}"
        else:
            return self._view_name

    @view_name.setter
    def view_name(self, value):
        self._view_name = value



import django.dispatch

plugin_loaded = django.dispatch.Signal(providing_args=["plugin"])
plugin_unloaded = django.dispatch.Signal(providing_args=["plugin"])

wamp_realm_created = django.dispatch.Signal()
wamp_realm_discarded = django.dispatch.Signal()



# Generated by Django 2.2 on 2019-04-25 13:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("unplugged", "0003_simpleadminplugin_display_name")]

    operations = [
        migrations.AlterField(
            model_name="simpleadminplugin",
            name="display_name",
            field=models.CharField(
                blank=True, help_text="Human readable name", max_length=200, null=True
            ),
        ),
        migrations.AlterField(
            model_name="simpleadminplugin",
            name="name",
            field=models.CharField(
                blank=True,
                help_text="Slugified name, used for actual name (and be able to find the plugin again)",
                max_length=200,
                null=True,
            ),
        ),
    ]



import re

from django.db import models
from jsonfield import JSONField

from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.interval import IntervalTrigger

from .plugin import Plugin


class FailedToParseScheduleException(Exception):
    pass


def parse_schedule_trigger(method, config):
    config = re.split(r" +", config)
    if method == "cron":
        fields = [
            "second",
            "minute",
            "hour",
            "day_of_week",
            "week",
            "day",
            "month",
            "year",
        ]
        try:
            kwargs = dict(zip(fields, config))
            return CronTrigger(**kwargs)
        except ValueError as e:
            raise FailedToParseScheduleException(*e.args)

    elif method == "interval":
        fields = ["weeks", "days", "hours", "minutes", "seconds"]
        try:
            kwargs = dict([c.split("=") for c in config])
            kwargs = {k: int(v) for (k, v) in kwargs.items()}
        except ValueError:
            raise FailedToParseScheduleException("Failed to parse config")

        unknown_keys = set(kwargs.keys()) - set(fields)
        if unknown_keys:
            raise FailedToParseScheduleException(
                f"Unknown keys found in config: {', '.join(unknown_keys)}"
            )

        try:
            return IntervalTrigger(**kwargs)
        except ValueError as e:
            raise FailedToParseScheduleException(*e.args)

    else:
        raise FailedToParseScheduleException(f"Unknown method: {method}")


class ScheduleManager(models.Manager):
    def ensure_schedule(
        self, method, method_config, command, kwargs, plugin, plugin_unique_id
    ):
        if not isinstance(plugin, Plugin):
            plugin = plugin._plugin_obj

        existing_schedules = self.filter(
            plugin=plugin, plugin_unique_id=plugin_unique_id
        )
        if existing_schedules:
            existing_schedule = existing_schedules[0]
            changed = False

            if existing_schedule.method != method:
                existing_schedule.method = method
                changed = True

            if existing_schedule.method_config != method_config:
                existing_schedule.method_config = method_config
                changed = True

            if existing_schedule.command != command:
                existing_schedule.command = command
                changed = True

            if existing_schedule.kwargs != kwargs:
                existing_schedule.kwargs = kwargs
                changed = True

            if not existing_schedule.enabled:
                existing_schedule.enabled = True
                changed = True

            if changed:
                existing_schedule.save()
        else:
            self.create(
                method=method,
                method_config=method_config,
                command=command,
                kwargs=kwargs,
                plugin=plugin,
                plugin_unique_id=plugin_unique_id,
            )


class Schedule(models.Model):
    METHOD_CRON = "cron"
    METHOD_INTERVAL = "interval"
    method = models.CharField(
        max_length=10, choices=((METHOD_CRON, "Cron"), (METHOD_INTERVAL, "Interval"))
    )
    method_config = models.CharField(max_length=500)

    command = models.CharField(max_length=200)
    kwargs = JSONField(blank=True)

    plugin = models.ForeignKey(Plugin, on_delete=models.CASCADE)
    plugin_unique_id = models.CharField(
        max_length=100, db_index=True, null=True, blank=True
    )

    enabled = models.BooleanField(default=False)

    objects = ScheduleManager()

    def get_trigger(self):
        return parse_schedule_trigger(self.method, self.method_config)



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "unplugged.services.api"
    verbose_name = "API Service"
    label = "services_api"

    def ready(self):
        from .handler import APIServicePlugin  # NOQA



import copy
import logging

from django.db import models
from django.utils.text import slugify
from jsonfield import JSONField

from ...baseplugin import RelatedPluginField
from ...models import Plugin
from ...pluginhandler import pluginhandler
from ...schema import Schema, fields

logger = logging.getLogger(__name__)


class NameAlreadyInUseException(Exception):
    pass


def key_name_in_use(plugin, modify_key, name):
    names = set([item["name"] for item in plugin.config.get(modify_key, [])])
    return name in names or slugify(name) in names


class SimpleAdminTemplate(models.Model):
    display_name = models.CharField(max_length=200)
    description = models.TextField(default="No description found")
    template_id = models.CharField(max_length=100)
    plugin_type = models.CharField(max_length=100)
    plugin_name = models.CharField(max_length=100)
    modify_key = models.CharField(max_length=40, null=True, blank=True)
    template = JSONField()

    UPDATE_METHOD_FULL = "full"
    UPDATE_METHOD_MODIFY_KEY = "modify_key"
    UPDATE_METHOD_CHOICES = (
        (UPDATE_METHOD_FULL, "Full"),
        (UPDATE_METHOD_MODIFY_KEY, "Modify Key"),
    )
    update_method = models.CharField(max_length=100, choices=UPDATE_METHOD_CHOICES)
    automatic_created = models.BooleanField(default=False)

    def __str__(self):
        return f"SimpleAdminTemplate {self.display_name}"

    class Meta:
        verbose_name_plural = "Simple Admin Templates"
        unique_together = (("template_id", "plugin_type", "plugin_name"),)
        db_table = "services_admin_simpleadmin_template"

    @property
    def config_schema(self):
        plugin = pluginhandler.get_plugin(self.plugin_type, self.plugin_name)
        if not plugin:
            return Schema

        schema = plugin.config_schema
        if self.update_method == self.UPDATE_METHOD_MODIFY_KEY:
            schema = schema._declared_fields[self.modify_key].nested

        def is_userinput(template, field_key):
            return (
                isinstance(template[field_key], dict)
                and template[field_key].get("simpleAdminMethod") == "userInput"
            )

        def is_injectable(template, field_key):
            return (
                isinstance(template[field_key], dict)
                and template[field_key].get("simpleAdminMethod") == "injectablePlugin"
            )

        def create_schema(schema, template, root=False, hide_fields=None):
            schema_keys = {}
            added_fields = False

            ui_schema = {"ui:order": []}
            existing_ui_schema = getattr(schema.Meta, "ui_schema", {})

            if template == True:
                ui_schema.update(existing_ui_schema)

            ui_schema["ui:order"] = list(ui_schema["ui:order"])

            for key, field in schema._declared_fields.items():
                if template != True and key not in template:
                    continue

                if hide_fields is not None:
                    if key not in hide_fields:
                        schema_keys[key] = field
                elif template == True:
                    schema_keys[key] = field
                elif is_userinput(template, key):
                    nested_hide_fields = template[key].get("hideFields")
                    new_field = copy.copy(field)
                    if nested_hide_fields:
                        nested_schema, nested_added_fields = create_schema(
                            field.nested, True, hide_fields=nested_hide_fields
                        )
                        new_field.nested = nested_schema
                    new_field.required = template[key].get("required", field.required)
                    schema_keys[key] = new_field
                    added_fields = True
                elif isinstance(field, fields.Nested):
                    nested_schema, nested_added_fields = create_schema(
                        field.nested, template[key]
                    )
                    if nested_added_fields:
                        new_field = copy.copy(field)
                        new_field.nested = nested_schema
                        schema_keys[key] = new_field
                        added_fields = True

            if isinstance(template, dict):
                for key, template_field in template.items():
                    if is_injectable(template, key):
                        plugin_type = pluginhandler.get_plugin_type(
                            template_field["plugin_type"]
                        )
                        traits = template_field.get("traits", [])
                        schema_keys[key] = RelatedPluginField(
                            required=True, plugin_type=plugin_type, traits=traits
                        )

            if (
                root
                and self.update_method == self.UPDATE_METHOD_FULL
                and "display_name" not in schema_keys
            ):
                # schema_keys['name'] = fields.String(required=True, pattern=r'^[a-z0-9]+(?:[-_][a-z0-9]+)*$', ui_schema={'ui:title': 'Name'})
                schema_keys["display_name"] = fields.String(
                    required=True,
                    default="",
                    ui_schema={"ui:title": "Name"},
                )
                ui_schema["ui:order"].insert(0, "display_name")

            for field_name in schema_keys.keys():
                if field_name not in ui_schema["ui:order"]:
                    ui_schema["ui:order"].append(field_name)

            existing_ui_order = ["display_name"] + existing_ui_schema.get(
                "ui:order", []
            )

            def get_index(f):
                try:
                    return existing_ui_order.index(f)
                except ValueError:
                    return 100

            ui_schema["ui:order"].sort(key=get_index)

            if root and self.update_method == self.UPDATE_METHOD_MODIFY_KEY:
                schema_keys[self.modify_key] = RelatedPluginField(
                    required=True, plugin_name=plugin, ui_schema={"ui:title": "Items"}
                )
                ui_schema["ui:order"].append(self.modify_key)

            schema_keys["Meta"] = type("Meta", (), {"ui_schema": ui_schema})

            schema_cls = type(schema.__name__, (Schema,), schema_keys)
            if root:
                return schema_cls
            else:
                return schema_cls, added_fields

        return create_schema(schema, self.template, root=True)

    def add_plugin(self, name, items=None):
        if self.update_method == self.UPDATE_METHOD_FULL:
            try:
                Plugin.objects.get(plugin_type=self.plugin_type, name=name)
                raise NameAlreadyInUseException(f"Name {name} is already used")
            except Plugin.DoesNotExist:
                plugin = Plugin.objects.create(
                    plugin_type=self.plugin_type,
                    plugin_name=self.plugin_name,
                    name=name,
                    enabled=True,
                )
        elif self.update_method == self.UPDATE_METHOD_MODIFY_KEY:
            plugins = Plugin.objects.filter(
                plugin_type=self.plugin_type, plugin_name=self.plugin_name
            ).order_by("id")

            if items:
                plugins = plugins.filter(pk=items)

            if not plugins:
                return None

            enabled_plugins = plugins.filter(enabled=True)

            if enabled_plugins:
                plugin = enabled_plugins[0]
            else:
                plugin = plugins[0]

            if key_name_in_use(plugin, self.modify_key, name):
                raise NameAlreadyInUseException(f"Name {name} is already used")
        else:
            raise Exception(f"Unknown update method {self.update_method}")

        sap = SimpleAdminPlugin.objects.create(template=self, name=name, plugin=plugin)

        return sap

    def get_injectable_plugin_fields(self):
        fields = []
        if isinstance(self.template, dict):
            for k, v in self.template.items():
                if not isinstance(v, dict):
                    continue

                if v.get("simpleAdminMethod") == "injectablePlugin":
                    fields.append(k)

        return fields


class SimpleAdminPlugin(models.Model):
    template = models.ForeignKey(SimpleAdminTemplate, on_delete=models.deletion.CASCADE)
    plugin = models.ForeignKey(Plugin, on_delete=models.deletion.CASCADE)
    auto_created_plugins = models.ManyToManyField(
        Plugin, related_name="auto_created_plugins", blank=True
    )
    name = models.CharField(
        max_length=200,
        blank=True,
        null=True,
        help_text="Slugified name, used for actual name (and be able to find the plugin again)",
    )
    display_name = models.CharField(
        max_length=200, blank=True, null=True, help_text="Human readable name"
    )
    config = JSONField(default=dict, blank=True)
    priority = models.SmallIntegerField(default=1)

    class Meta:
        verbose_name_plural = "Simple Admin Plugin"
        db_table = "services_admin_simpleadmin_plugin"
        ordering = ["priority"]

    def delete_plugin(self):
        for auto_created_plugin in self.auto_created_plugins.all():
            logger.debug(
                "Deleting auto-created plugin %r from %r"
                % (auto_created_plugin, self.plugin)
            )
            auto_created_plugin.remove_plugin()
            auto_created_plugin.delete()

        if self.template.update_method == self.template.UPDATE_METHOD_FULL:
            logger.debug(f"Removing plugin {self.plugin} completely with plugin {self}")
            self.plugin.remove_plugin()
            self.plugin.delete()
        elif self.template.update_method == self.template.UPDATE_METHOD_MODIFY_KEY:
            logger.debug(f"Removing a key plugin from {self.plugin} with plugin {self}")
            found_item = False
            for i, item in enumerate(
                self.plugin.config.get(self.template.modify_key, [])
            ):
                if (
                    item["name"] == self.name
                    or item.get("_simpleadminplugin_id") == self.pk
                ):
                    found_item = True
                    break

            if found_item:
                self.plugin.config[self.template.modify_key].pop(i)
                self.plugin.save()

    def check_in_use_by(self):
        if self.plugin.is_plugin_loaded():
            plugin_obj = self.plugin.get_plugin()
            return plugin_obj._related_plugins

        return []

    def reload_plugin(self):
        self.plugin.reload_plugin()

    def update_plugin(
        self, form_data
    ):  # TODO: maybe make it choose a new name if already in use?
        """
        Update the actual plugin with the new form_data.
        """
        filled_form_data = self.fill_form_data(form_data)
        filled_form_data["_simpleadminplugin_id"] = self.pk
        name = form_data.get("name") or filled_form_data.get("name") or self.name
        display_name = (
            form_data.get("display_name")
            or filled_form_data.get("display_name")
            or self.display_name
        )
        assert name

        if name != self.name:
            if self.template.update_method == self.template.UPDATE_METHOD_FULL:
                try:
                    Plugin.objects.get(plugin_type=self.plugin.plugin_type, name=name)
                    raise NameAlreadyInUseException(
                        f"Unable to rename because name {name} is already used"
                    )
                except Plugin.DoesNotExist:
                    pass
            elif self.template.update_method == self.template.UPDATE_METHOD_MODIFY_KEY:
                if key_name_in_use(self.plugin, self.template.modify_key, name):
                    raise NameAlreadyInUseException(
                        f"Unable to save because name {name} is already used"
                    )

        plugin_cls = pluginhandler.get_plugin(
            self.template.plugin_type, self.template.plugin_name
        )
        schema = plugin_cls.config_schema
        if self.template.update_method == self.template.UPDATE_METHOD_MODIFY_KEY:
            schema = schema._declared_fields[self.template.modify_key].nested
        filled_form_data.update(schema().dump(filled_form_data))

        should_reload = False
        if self.template.update_method == self.template.UPDATE_METHOD_FULL:
            logger.debug(f"Doing a full fill of form data to {self.plugin} from {self}")
            self.plugin.config = filled_form_data
            self.plugin.name = name
            should_reload = True
        elif self.template.update_method == self.template.UPDATE_METHOD_MODIFY_KEY:
            logger.debug(
                f"Doing a partial fill of form data to {self.plugin} from {self}"
            )

            if self.plugin.config.get(self.template.modify_key):
                updated_item = False
                for item in self.plugin.config[self.template.modify_key]:
                    if (
                        item["name"] == self.name
                        or item.get("_simpleadminplugin_id") == self.pk
                    ):
                        item.clear()
                        item.update(filled_form_data)
                        updated_item = True
                        break

                if not updated_item:
                    self.plugin.config[self.template.modify_key].append(
                        filled_form_data
                    )
            else:
                self.plugin.config[self.template.modify_key] = [filled_form_data]

        for key in self.template.get_injectable_plugin_fields():
            if key in form_data:
                self.config[key] = form_data[key]

        self.plugin.enabled = True
        self.plugin.save()

        if should_reload:
            self.reload_plugin()

        self.name = name
        if display_name:
            self.display_name = display_name

    def fill_form_data(self, form_data):
        """
        Fills the partial form with auto-completed fields from the template.
        """

        def create_plugin(template):
            if self.pk is not None:
                # full_name = '%s_%s' % (self.pk, template['name'])
                # vars = {'name': full_name, 'id': self.pk}
                name = template["name"] % {"name": self.pk}
                config = {
                    k: v % {"name": self.pk}
                    for (k, v) in template.get("config", {}).items()
                }
                plugin, created = Plugin.objects.get_or_create(
                    name=name,
                    plugin_type=template["plugin_type"],
                    plugin_name=template["plugin_name"],
                    config=config,
                    defaults={"enabled": True},
                )
                if created:
                    self.auto_created_plugins.add(plugin)
                else:
                    pass  # TODO: make sure it's not autocreated
                return plugin.pk
            else:
                return None

        injectable_plugins = {}

        def fill_data(template, source_form_data=None, key=None):
            if isinstance(template, dict):
                method = template.get("simpleAdminMethod")
                if method:
                    if method == "userInput":
                        return source_form_data
                    elif method == "lookupPlugin":
                        plugins = Plugin.objects.filter(
                            plugin_type=template["plugin_type"],
                            name=template["name"],
                            enabled=True,
                        )
                        if plugins:
                            return plugins[0].pk
                        else:
                            logger.warning(
                                f"Unable to find plugin plugin_type:{template['plugin_type']} name:{template['name']}"
                            )
                            return None
                    elif method == "lookupAllPlugins":
                        return Plugin.objects.filter(
                            plugin_type=template["plugin_type"], enabled=True
                        ).values_list("pk", flat=True)
                    elif method == "createPlugin":
                        return create_plugin(template)
                    elif method == "createPluginPriority":
                        for plugin_template in template["pluginPriorities"]:
                            if not pluginhandler.get_plugin(
                                plugin_template["plugin_type"],
                                plugin_template["plugin_name"],
                            ):
                                continue
                            return create_plugin(plugin_template)
                    elif method == "slugify":
                        return slugify(form_data[template["source"]])
                    elif method == "injectablePlugin":
                        logger.debug("We got data: %r" % (source_form_data,))
                        injectable_plugins[key] = source_form_data
                    elif method == "injectedPlugin":
                        return injectable_plugins[template["id"]]
                else:
                    retval = {}
                    for k, v in template.items():
                        if isinstance(v, list):
                            result = []
                            for vv in v:
                                filled_value = fill_data(vv, key=k)
                                if filled_value is not None:
                                    result.append(filled_value)
                            retval[k] = result
                        else:
                            source_value = None
                            if (
                                isinstance(source_form_data, dict)
                                and source_form_data.get(k) is not None
                            ):
                                source_value = source_form_data[k]

                            filled_value = fill_data(v, source_value, key=k)
                            if filled_value is not None:
                                retval[k] = filled_value

                    return retval
            else:
                return template

        if self.template.template == True:
            return form_data
        else:
            template = self.template.template
            if (
                isinstance(template, dict)
                and "name" not in template
                and "display_name" in template
            ):
                template["name"] = {
                    "simpleAdminMethod": "slugify",
                    "source": "display_name",
                }
            return fill_data(self.template.template, form_data)

    def pull_form_data(self):
        schema = self.template.config_schema
        config = dict(self.plugin.config)
        config["name"] = self.plugin.name
        if "display_name" not in config:
            config["display_name"] = self.get_display_name()

        data = {}

        if self.template.update_method == self.template.UPDATE_METHOD_FULL:
            data = schema().dump(config)
        elif self.template.update_method == self.template.UPDATE_METHOD_MODIFY_KEY:
            for item in config.get(self.template.modify_key, []):
                if item["name"] == self.name:
                    item = dict(item)
                    item[self.template.modify_key] = self.plugin.pk
                    data = schema().dump(item)

        for key in self.template.get_injectable_plugin_fields():
            if key in self.config:
                data[key] = self.config.get(key)

        return data

    def get_display_name(self):
        config = self.plugin.config
        if self.template.update_method == self.template.UPDATE_METHOD_MODIFY_KEY:
            for item in config.get(self.template.modify_key, []):
                if item["name"] == self.name:
                    config = item
                    break
            else:
                config = {}

        return (
            self.display_name
            or config.get("display_name")
            or self.name
            or "No name found"
        )

    def __str__(self):
        return f"SimpleAdminPlugin {self.name}"



import logging

from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls import NoReverseMatch, reverse
from django.utils.text import slugify
from marshmallow import INCLUDE
from rest_framework import permissions, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.views import APIView

from ....jsonapi import JSONAPIObject, JSONAPIRoot
from ....jsonschema import dump_ui_schema
from ....libs.marshmallow_jsonschema import JSONSchema
from ....schema import ValidationError
from ..models import NameAlreadyInUseException, SimpleAdminPlugin, SimpleAdminTemplate
from .shared import ADMIN_RENDERER_CLASSES, ServiceAwareHyperlinkedIdentityField

logger = logging.getLogger(__name__)


class SimpleAdminTemplateSerializer(serializers.HyperlinkedModelSerializer):
    schema = serializers.SerializerMethodField(source="template")
    ui_schema = serializers.SerializerMethodField(source="template")

    class Meta:
        model = SimpleAdminTemplate
        fields = (
            "display_name",
            "description",
            "template_id",
            "plugin_type",
            "plugin_name",
            "schema",
            "ui_schema",
        )

    class JSONAPIMeta:
        resource_name = "simpleadmin_template"

    def get_schema(self, obj):
        json_schema = JSONSchema()
        return json_schema.dump(obj.config_schema())

    def get_ui_schema(self, obj):
        schema = obj.config_schema()
        return dump_ui_schema(schema)


class SimpleAdminTemplateModelView(viewsets.ModelViewSet):
    """
    List of all possible templates we can create listing derivatives from.
    """

    queryset = SimpleAdminTemplate.objects.all()
    serializer_class = SimpleAdminTemplateSerializer
    renderer_classes = ADMIN_RENDERER_CLASSES
    permission_classes = (permissions.IsAdminUser,)
    pagination_class = None

    service = None


class SimpleAdminPluginSerializer(serializers.HyperlinkedModelSerializer):
    url = ServiceAwareHyperlinkedIdentityField(view_name="simpleadmin_plugin-detail")
    template = SimpleAdminTemplateSerializer()
    plugin_type = serializers.CharField(read_only=True, source="plugin.plugin_type")
    plugin_name = serializers.CharField(read_only=True, source="plugin.plugin_name")
    config = serializers.SerializerMethodField()
    display_name = serializers.SerializerMethodField()

    class Meta:
        model = SimpleAdminPlugin
        fields = (
            "name",
            "template",
            "template_id",
            "plugin_id",
            "plugin_type",
            "plugin_name",
            "url",
            "config",
            "display_name",
        )

    class JSONAPIMeta:
        resource_name = "simpleadmin_plugin"

    def get_config(self, obj):
        return obj.pull_form_data()

    def get_display_name(self, obj):
        return obj.get_display_name()


class SimpleAdminPluginCreateSerializer(serializers.ModelSerializer):
    config = serializers.DictField(write_only=True)  # verify it matches schema
    template_id = serializers.IntegerField()
    plugin_type = serializers.CharField()
    plugin_name = serializers.CharField()

    class Meta:
        model = SimpleAdminPlugin
        fields = ("config", "template_id", "plugin_type", "plugin_name")

    class JSONAPIMeta:
        resource_name = "simpleadmin_plugin"


class SimpleAdminPluginPrioritySerializer(serializers.Serializer):
    priorities = serializers.ListField(child=serializers.IntegerField())


class SimpleAdminPluginModelView(viewsets.ModelViewSet):
    """
    What is currently controlled by simpleadmin and where?
    """

    queryset = SimpleAdminPlugin.objects.all().order_by("priority")
    serializer_class = SimpleAdminPluginSerializer
    renderer_classes = ADMIN_RENDERER_CLASSES
    permission_classes = (permissions.IsAdminUser,)
    pagination_class = None

    service = None

    @action(methods=["post"], detail=False)
    def set_priorities(self, request):
        serializer = SimpleAdminPluginPrioritySerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        plugins = set()
        priorities = {}
        for i, plugin_id in enumerate(serializer.validated_data["priorities"]):
            sap = SimpleAdminPlugin.objects.get(id=plugin_id)
            sap.priority = i
            sap.save(update_fields=["priority"])

            plugins.add(sap.plugin)
            priorities[(sap.plugin, sap.name)] = i
            priorities[(sap.plugin, sap.pk)] = i

        for plugin in plugins:

            def get_priority(p):
                k = (plugin, p["name"])
                if k in priorities:
                    return priorities[k]

                k = (plugin, p["_simpleadminplugin_id"])
                if k in priorities:
                    return priorities[k]

                return 100

            plugin.config[sap.template.modify_key] = sorted(
                plugin.config[sap.template.modify_key], key=get_priority
            )
            plugin.save()

        return Response({"status": "success", "message": "Priority set"})

    @action(methods=["get"], detail=True)
    def in_use_by(self, request, pk=None):
        plugin = self.get_object()

        return Response(
            [
                {
                    "name": p.name,
                    "plugin_name": p.plugin_name,
                    "plugin_type": p.plugin_type,
                }
                for p in plugin.check_in_use_by()
            ]
        )

    def get_serializer_class(self):
        if self.action in ["create", "update"]:
            return SimpleAdminPluginCreateSerializer
        else:
            return super(SimpleAdminPluginModelView, self).get_serializer_class()

    def generate_name(self, serializer):
        name = serializer.initial_data["config"].get("name")
        if not name:
            name = slugify(serializer.initial_data["config"].get("display_name"))
        return name

    def perform_create(self, serializer):
        template = get_object_or_404(
            SimpleAdminTemplate, pk=serializer.initial_data["template_id"]
        )
        try:
            _ = template.config_schema().load(
                serializer.initial_data["config"], unknown=INCLUDE
            )
        except ValidationError as err:
            return Response(err.messages, status=status.HTTP_400_BAD_REQUEST)

        name = self.generate_name(serializer)

        try:
            sap = template.add_plugin(
                name, serializer.initial_data["config"].get(template.modify_key)
            )
        except NameAlreadyInUseException as e:
            return Response({"name": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)

        if not sap:
            raise Http404

        try:
            sap.update_plugin(serializer.initial_data["config"])
        except NameAlreadyInUseException as e:
            return Response({"name": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
        sap.save()

        return sap

    def perform_update(self, serializer):
        sap = serializer.instance
        try:
            _ = sap.template.config_schema().load(
                serializer.initial_data["config"], unknown=INCLUDE
            )
        except ValidationError as err:
            return Response(err.messages, status=status.HTTP_400_BAD_REQUEST)

        name = self.generate_name(serializer)
        if sap.name != name:
            logger.debug("Trying to rename %s to %s" % (sap.name, name))
            serializer.initial_data["config"]["name"] = name

        try:
            sap.update_plugin(serializer.initial_data["config"])
        except NameAlreadyInUseException as e:
            return Response({"name": [str(e)]}, status=status.HTTP_400_BAD_REQUEST)
        sap.save()

    def perform_destroy(self, instance):
        instance.delete_plugin()
        instance.delete()

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)

        r = self.perform_update(serializer)
        if isinstance(r, Response):
            return r

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        serializer_class = super(
            SimpleAdminPluginModelView, self
        ).get_serializer_class()
        return Response(
            serializer_class(
                instance=instance, context={"request": request, "view": self}
            ).data
        )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = self.perform_create(serializer)
        if isinstance(obj, Response):
            return obj
        headers = self.get_success_headers(serializer.data)
        serializer_class = super(
            SimpleAdminPluginModelView, self
        ).get_serializer_class()
        return Response(
            serializer_class(
                instance=obj, context={"request": request, "view": self}
            ).data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )


class ShowAdminUrlsView(APIView):
    permission_classes = (permissions.IsAdminUser,)

    router = None
    urls = None
    service = None

    def get(self, request):
        root = JSONAPIRoot()

        for url in self.urls:
            if url.name.endswith("-detail"):
                continue

            resource_name = url.callback.cls.serializer_class.JSONAPIMeta.resource_name
            try:
                view_url = reverse("unplugged:%s:%s" % (self.service.name, url.name))
            except NoReverseMatch:
                continue

            if url.name.endswith("-list"):
                view_id = resource_name
            else:
                view_id = "%s_%s" % (
                    resource_name,
                    url.name.split("-", 1)[1].replace("-", "_"),
                )

            links = {"self": request.build_absolute_uri(view_url)}
            plugin_type = "admin_%s" % (view_id,)
            obj = JSONAPIObject(plugin_type, resource_name, links=links)
            root.append(obj)

        return Response(root.serialize(request))



from .log import LogModelView
from .plugin import PluginBaseListView, PluginModelView
from .scheduler import ScheduleModelView

__all__ = ["LogModelView", "PluginBaseListView", "PluginModelView", "ScheduleModelView"]



import datetime
import decimal
import uuid
from inspect import isclass

from marshmallow import Schema, fields, missing, validate
from marshmallow.class_registry import get_class
from marshmallow.decorators import post_dump

from .compat import basestring, binary_type, dot_data_backwards_compatible, text_type
from .validation import handle_length, handle_one_of, handle_range

__all__ = ("JSONSchema",)


TYPE_MAP = {
    dict: {"type": "object"},
    list: {"type": "array"},
    datetime.time: {"type": "string", "format": "time"},
    datetime.timedelta: {
        # TODO explore using 'range'?
        "type": "string"
    },
    datetime.datetime: {"type": "string", "format": "date-time"},
    datetime.date: {"type": "string", "format": "date"},
    uuid.UUID: {"type": "string", "format": "uuid"},
    text_type: {"type": "string"},
    binary_type: {"type": "string"},
    decimal.Decimal: {"type": "number", "format": "decimal"},
    set: {"type": "array"},
    tuple: {"type": "array"},
    float: {"type": "number", "format": "float"},
    int: {"type": "number", "format": "integer"},
    bool: {"type": "boolean"},
}


FIELD_VALIDATORS = {
    validate.Length: handle_length,
    validate.OneOf: handle_one_of,
    validate.Range: handle_range,
}


class JSONSchema(Schema):
    """Converts to JSONSchema as defined by http://json-schema.org/."""

    properties = fields.Method("get_properties")
    type = fields.Constant("object")
    required = fields.Method("get_required")

    def __init__(self, *args, **kwargs):
        """Setup internal cache of nested fields, to prevent recursion."""
        self._nested_schema_classes = {}
        self.nested = kwargs.pop("nested", False)
        super(JSONSchema, self).__init__(*args, **kwargs)

    def _get_default_mapping(self, obj):
        """Return default mapping if there are no special needs."""
        mapping = {v: k for k, v in obj.TYPE_MAPPING.items()}
        mapping.update(
            {
                fields.Email: text_type,
                fields.Dict: dict,
                fields.Url: text_type,
                fields.List: list,
                fields.Nested: "_from_nested_schema",
            }
        )
        return mapping

    def get_properties(self, obj):
        """Fill out properties field."""
        properties = {}

        for field_name, field in sorted(obj.fields.items()):
            schema = self._get_schema_for_field(obj, field)
            properties[field.name] = schema

        return properties

    def get_required(self, obj):
        """Fill out required field."""
        required = []

        for field_name, field in sorted(obj.fields.items()):
            if field.required:
                required.append(field.name)

        return required or missing

    def _from_python_type(self, obj, field, pytype):
        """Get schema definition from python type."""
        json_schema = {"title": field.attribute or field.name}

        for key, val in TYPE_MAP[pytype].items():
            json_schema[key] = val

        if field.dump_only:
            json_schema["readonly"] = True

        if field.default is not missing:
            if isinstance(field.default, type):
                json_schema["default"] = field.default()
            else:
                json_schema["default"] = field.default

        # NOTE: doubled up to maintain backwards compatibility
        metadata = field.metadata.get("metadata", {})
        metadata.update(field.metadata)

        for md_key, md_val in metadata.items():
            if md_key == "metadata":
                continue
            json_schema[md_key] = md_val

        if isinstance(field, fields.List):
            json_schema["items"] = self._get_schema_for_field(obj, field.inner)
        return json_schema

    def _get_schema_for_field(self, obj, field):
        """Get schema and validators for field."""
        mapping = self._get_default_mapping(obj)
        if hasattr(field, "_jsonschema_type_mapping"):
            schema = field._jsonschema_type_mapping()
        elif "_jsonschema_type_mapping" in field.metadata:
            schema = field.metadata["_jsonschema_type_mapping"]
        elif field.__class__ in mapping:
            pytype = mapping[field.__class__]
            if isinstance(pytype, basestring):
                schema = getattr(self, pytype)(obj, field)
            else:
                schema = self._from_python_type(obj, field, pytype)
        else:
            raise ValueError("unsupported field type %s" % field)

        # Apply any and all validators that field may have
        for validator in field.validators:
            if validator.__class__ in FIELD_VALIDATORS:
                schema = FIELD_VALIDATORS[validator.__class__](
                    schema, field, validator, obj
                )
        return schema

    def _from_nested_schema(self, obj, field):
        if isinstance(field.nested, basestring):
            nested = get_class(field.nested)
        else:
            nested = field.nested

        if isclass(nested) and issubclass(nested, Schema):
            only = field.only
            exclude = field.exclude
            nested_instance = nested(only=only, exclude=exclude)
        else:
            nested_instance = nested

        wrapped_nested = self.__class__(nested=True)
        wrapped_dumped = wrapped_nested.dump(nested_instance)

        # NOTE: doubled up to maintain backwards compatibility
        metadata = field.metadata.get("metadata", {})
        metadata.update(field.metadata)

        for md_key, md_val in metadata.items():
            if md_key == "metadata":
                continue
            wrapped_dumped[md_key] = md_val

        if field.many:
            wrapped_dumped = {
                "type": "array" if field.required else ["array", "null"],
                "items": wrapped_dumped,
            }

        return wrapped_dumped

    def _from_nested_schema_old(self, obj, field):
        """Support nested field."""
        if isinstance(field.nested, basestring):
            nested = get_class(field.nested)
        else:
            nested = field.nested

        if isclass(nested) and issubclass(nested, Schema):
            name = nested.__name__
            only = field.only
            exclude = field.exclude
            nested_instance = nested(only=only, exclude=exclude)
        else:
            name = nested.__class__.__name__
            nested_instance = nested

        outer_name = obj.__class__.__name__
        # If this is not a schema we've seen, and it's not this schema (checking this for recursive schemas),
        # put it in our list of schema defs
        if name not in self._nested_schema_classes and name != outer_name:
            wrapped_nested = self.__class__(nested=True)
            wrapped_dumped = wrapped_nested.dump(nested_instance)

            # Handle change in return value type between Marshmallow
            # versions 2 and 3.
            self._nested_schema_classes[name] = dot_data_backwards_compatible(
                wrapped_dumped
            )

            self._nested_schema_classes.update(wrapped_nested._nested_schema_classes)

        # and the schema is just a reference to the def
        schema = {"type": "object", "$ref": "#/definitions/{}".format(name)}

        # NOTE: doubled up to maintain backwards compatibility
        metadata = field.metadata.get("metadata", {})
        metadata.update(field.metadata)

        for md_key, md_val in metadata.items():
            if md_key == "metadata":
                continue
            schema[md_key] = md_val

        if field.many:
            schema = {
                "type": "array" if field.required else ["array", "null"],
                "items": schema,
            }

        return schema

    def dump(self, obj, **kwargs):
        """Take obj for later use: using class name to namespace definition."""
        self.obj = obj
        return super(JSONSchema, self).dump(obj, **kwargs)

    # @post_dump
    # def wrap(self, data, **_):
    #     """Wrap this with the root schema definitions."""
    #     if self.nested:  # no need to wrap, will be in outer defs
    #         return data

    #     name = self.obj.__class__.__name__
    #     self._nested_schema_classes[name] = data
    #     root = {
    #         "definitions": self._nested_schema_classes,
    #         "$ref": "#/definitions/{name}".format(name=name),
    #     }
    #     return root



import logging

from wampyre.transports.django import WAMPRouter

from ...schema import Schema
from ...plugins import ServicePlugin

logger = logging.getLogger(__name__)


class WAMPServicePlugin(ServicePlugin):
    plugin_name = "wamp"
    config_schema = Schema

    def get_channels(self):
        return [
            (
                "",
                WAMPRouter,
                {
                    "realm_authenticator": self.realm_authenticator,
                    "guard": self.wamp_guard,
                },
            )
        ]

    def realm_authenticator(user, realm):
        if not user or not user.is_authenticated():
            logger.info(f"User not authenticated for realm {realm}")
            return False

        if realm != "unplugged":
            logger.info(f"Someone tried to connect to realm {realm}")
            return False

        return True

    def wamp_guard(user, method, uri):
        if not user:
            return False

        if uri.startswith("user."):
            if user.is_staff:
                return True

            user_id = uri.split(".")[1]

            if str(user.pk) == user_id:
                return True

            return False
        elif uri.startswith("plugin."):
            uri = uri.split(".")
            if len(uri) < 3:
                return False

            plugin_type, name = uri.split(".")[1:3]
            if user.has_perm(f"unplugged.{plugin_type}.{name}"):
                return True

        return False



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-12-05 20:25
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("services_user", "0001_initial")]

    operations = [
        migrations.AlterUniqueTogether(name="userinfo", unique_together=set([])),
        migrations.RemoveField(model_name="userinfo", name="user"),
        migrations.DeleteModel(name="Userinfo"),
    ]



"""
Coordinates plugins and handles the register
"""

import logging

logger = logging.getLogger(__name__)


class UnknownPluginTypeException(Exception):
    pass


class NotASubclassException(Exception):
    pass


class PluginHandler:
    def __init__(self):
        self.plugin_types = {}
        self.plugins = {}

    def register_plugin_type(self, plugin_interface):
        """
        Register a new plugin type.
        """
        plugin_type = plugin_interface.plugin_type
        if plugin_type in self.plugin_types:
            logger.info(f"Trying to register duplicate plugin_type:{plugin_type}")
            return

        logger.info("Registering plugin type: %r" % (plugin_type,))
        self.plugin_types[plugin_type] = plugin_interface
        self.plugins[plugin_type] = {}

    def register_plugin(self, cls):
        """
        Register a new plugin, making it usable.
        """
        plugin_type = cls.plugin_type
        name = cls.plugin_name

        if not name:
            raise AttributeError("Missing name")

        if plugin_type not in self.plugin_types:
            logger.error(f"Unknown plugin type {plugin_type} for plugin {name}")
            raise UnknownPluginTypeException(
                f"{plugin_type} is not a known plugin type"
            )

        if not issubclass(cls, self.plugin_types[plugin_type]):
            logger.error(f"{name} is not a subclass of {plugin_type}")
            raise NotASubclassException(f"{name} is not a subclass of {plugin_type}")

        if name in self.plugins[plugin_type]:
            logger.warning(f"Double registering plugin {name} of type {plugin_type}")

        logger.info(f"Registering plugin {name} of type {plugin_type}")
        self.plugins[plugin_type][name] = cls

    def get_all_plugins(self):
        for plugin_type in self.plugins.keys():
            for plugin in self.get_plugins(plugin_type):
                yield plugin

    def get_plugins(self, plugin_type):
        return sorted(
            self.plugins[plugin_type].values(), key=lambda x: getattr(x, "priority", 0)
        )

    def get_plugin_names(self, plugin_type):
        return self.plugins[plugin_type].keys()

    def get_plugin(self, plugin_type, name):
        return self.plugins[plugin_type].get(name)

    def get_plugin_type(self, plugin_type):
        return self.plugins[plugin_type]


pluginhandler = PluginHandler()



import logging

from django.contrib.auth import authenticate, get_user_model, login, logout
from rest_framework import permissions, serializers, status
from rest_framework.authtoken.models import Token
from rest_framework.response import Response
from rest_framework.views import APIView

from ...jsonapi import JSONAPIObject, JSONAPIRoot

logger = logging.getLogger(__name__)


class AuthSerializer(serializers.Serializer):
    username = serializers.CharField(min_length=1, max_length=100)
    password = serializers.CharField(min_length=1, max_length=100)


class AuthView(APIView):
    service = None
    permission_classes = (permissions.AllowAny,)

    def post(self, request):
        serializer = AuthSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        UserModel = get_user_model()

        user = authenticate(
            username=serializer.data["username"], password=serializer.data["password"]
        )

        if user is not None:
            if user.is_active:
                logger.info("User %r logged in successfully" % (user,))
                login(request, user)
                return Response(
                    {"status": "success", "message": "logged in successfully"}
                )
            else:
                logger.info("User %r tried to login but is not active" % (user,))
                return Response({"status": "error", "message": "not authenticated"})
        elif not UserModel.objects.all().exists():
            logger.info("Creating new user as none exist")
            user = UserModel.objects.create_superuser(
                username=serializer.data["username"],
                password=serializer.data["password"],
                email=None,
            )
            login(request, user)
            return Response(
                {"status": "success", "message": "Created new user and logged in"}
            )
        else:
            logger.info("Login failed for username %s" % (serializer.data["username"],))
            return Response({"status": "error", "message": "not authenticated"})


class UserView(APIView):
    service = None
    permission_classes = (permissions.AllowAny,)

    def get(self, request):
        obj = JSONAPIObject("userinfo", request.user.username)

        obj["is_authenticated"] = bool(request.user.is_authenticated)

        if request.user.is_authenticated:
            obj["username"] = request.user.username
            obj["token"] = Token.objects.get_or_create(user=request.user)[0].key

        obj.links = {
            "self": request.build_absolute_uri("/%s/" % (self.service.name,)),
            "auth": request.build_absolute_uri("/%s/%s" % (self.service.name, "auth")),
            "logout": request.build_absolute_uri(
                "/%s/%s" % (self.service.name, "logout")
            ),
        }

        root = JSONAPIRoot()
        root.append(obj)

        return Response(root.serialize(request))


class LogoutView(APIView):
    service = None
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request):
        logout(request)
        return Response({"status": "success"})



# Generated by Django 2.2 on 2019-04-13 18:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("unplugged", "0002_simpleadminplugin_simpleadmintemplate")]

    operations = [
        migrations.AddField(
            model_name="simpleadminplugin",
            name="display_name",
            field=models.CharField(blank=True, max_length=200, null=True),
        )
    ]



import logging

from django.conf import settings
from django.conf.urls import include, url
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.urls import path
from django.utils.text import slugify
from rest_framework.permissions import BasePermission

from ..baseplugin import PluginBase
from ..routing import register_channels, unregister_channel
from ..urls import register_urlpatterns, unregister_urlpattern

logger = logging.getLogger(__name__)


class CascadingPermission(BasePermission):
    """
    Tries to check up the dependency tree to see
    if there is a service allowing this plugin.
    """

    def has_permission(self, request, view):
        if not request.user or not request.user.is_authenticated:
            return False

        checked = []

        def check_up(plugin):
            perm_name = "unplugged.%s.%s" % (plugin.plugin_type, plugin.name)
            checked.append(plugin)
            logger.debug(
                "Checking permission %r for user %r" % (perm_name, request.user)
            )
            if request.user.has_perm(perm_name):
                return True

            for sub_plugin in plugin._related_plugins:
                if sub_plugin in checked:
                    continue

                if check_up(sub_plugin):
                    return True

        return bool(check_up(view.service))


class DefaultPermission:
    IGNORE = "ignore"
    ALLOW = "allow"
    DENY = "deny"


class ServicePlugin(PluginBase):
    plugin_type = "service"
    mount_at_root = False
    default_permission = DefaultPermission.IGNORE

    def get_urls(self):
        """
        Returns urls to be registered.
        """
        return []

    def get_channels(self):
        """
        Returns channels to be registered.
        """
        return []

    def unload(self):
        unregister_urlpattern(r"^%s/" % (slugify(self.name),))
        unregister_channel(r"^%s/" % (slugify(self.name),))

    def ready(self):
        service_urls = self.get_urls()
        if service_urls:
            path_fragment = r"^%s/" % (slugify(self.name),)
            urls = [
                url(
                    path_fragment,
                    include((service_urls, self.name), namespace=self.name),
                )
            ]

            if self.mount_at_root:
                logger.info("Mounting %s at root" % (self.name,))
                urls.append(url("^/?$", include(service_urls)))

            register_urlpatterns(urls)

        channels = self.get_channels()
        if channels:
            path_fragment = r"%s/" % (slugify(self.name),)
            real_channels = []
            for channel in channels:
                if len(channel) == 2:
                    p, consumer = channel
                    kwargs = None
                elif len(channel) == 3:
                    p, consumer, kwargs = channel

                real_channels.append(path(path_fragment + p, consumer, kwargs=kwargs))
            register_channels(real_channels)


@receiver(
    post_save, sender=settings.AUTH_USER_MODEL, dispatch_uid="set_initial_permissions"
)
def set_initial_permissions(sender, instance, created, **kwargs):
    if not created:
        return

    from ..models import Plugin

    logger.info("Setting default permissions for user %r" % (instance,))
    for plugin in Plugin.objects.filter(enabled=True, plugin_type="service"):
        plugin_obj = plugin.get_plugin()
        if plugin_obj.default_permission == "allow":
            instance.user_permissions.add(plugin_obj._permission)



from rest_framework import permissions, serializers, viewsets
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework_json_api.renderers import JSONRenderer

from ..models import FailedToParseScheduleException, Schedule, parse_schedule_trigger

ADMIN_RENDERER_CLASSES = (JSONRenderer, BrowsableAPIRenderer)


class ScheduleSerializer(serializers.HyperlinkedModelSerializer):
    plugin_id = serializers.IntegerField()
    kwargs = serializers.DictField(default=dict)

    class Meta:
        model = Schedule
        fields = "__all__"
        read_only_fields = ["plugin"]

    class JSONAPIMeta:
        resource_name = "schedule"

    def validate(self, data):
        try:
            parse_schedule_trigger(data["method"], data["method_config"])
        except FailedToParseScheduleException as e:
            raise serializers.ValidationError(f"Failed to parse schedule: {e.args[0]}")

        return data


class ScheduleModelView(viewsets.ModelViewSet):
    queryset = Schedule.objects.all()
    serializer_class = ScheduleSerializer
    renderer_classes = ADMIN_RENDERER_CLASSES
    permission_classes = (permissions.IsAdminUser,)
    pagination_class = None
    filterset_fields = ("plugin",)



from .baseplugin import DjangoModelField, PluginBase, RelatedPluginField
from .commands import (
    Command,
    CommandBaseMeta,
    CommandSerializer,
    CommandViewMixin,
    command,
)
from .jsonapi import JSONAPIObject, JSONAPIRoot
from .jsonschema import dump_ui_schema
from .libs.marshmallow_jsonschema import JSONSchema
from .pluginhandler import pluginhandler
from .plugins import CascadingPermission, DefaultPermission, ServicePlugin
from .schema import Schema, fields
from .utils import deferToThreadWrapper

__all__ = [
    "DjangoModelField",
    "PluginBase",
    "RelatedPluginField",
    "Command",
    "CommandSerializer",
    "CommandViewMixin",
    "command",
    "CommandBaseMeta",
    "JSONAPIObject",
    "JSONAPIRoot",
    "dump_ui_schema",
    "pluginhandler",
    "ServicePlugin",
    "CascadingPermission",
    "DefaultPermission",
    "Schema",
    "fields",
    "JSONSchema",
    "deferToThreadWrapper",
]



import copy
import logging

from django.conf.urls import url
from rest_framework import routers

from ...schema import Schema
from ...pluginhandler import pluginhandler
from ...plugins import ServicePlugin
from .models import SimpleAdminPlugin, SimpleAdminTemplate
from .views import (
    LogModelView,
    PermissionModelView,
    PluginBaseListView,
    PluginModelView,
    ScheduleModelView,
    ShowAdminUrlsView,
    SimpleAdminPluginModelView,
    SimpleAdminTemplateModelView,
    UserModelView,
)

logger = logging.getLogger(__name__)


def register_simpleadmin_template(plugin_base, template):
    if template == True:
        description = "Default template for plugin"
        display_name = "Default"
        template_id = "default"
        simpleadmin_template = True
        modify_key = None
        update_method = SimpleAdminTemplate.UPDATE_METHOD_FULL
    else:
        description = template.get("description", "No description found")
        template_id = template["id"]
        display_name = template.get("display_name", template_id)
        simpleadmin_template = template["template"]
        update_method = template["update_method"]
        modify_key = template.get("modify_key")

    if update_method == SimpleAdminTemplate.UPDATE_METHOD_MODIFY_KEY and not modify_key:
        logger.warning(f"Bogus template {template_id}")
        return

    try:
        sat = SimpleAdminTemplate.objects.get(
            template_id=template_id,
            plugin_type=plugin_base.plugin_type,
            plugin_name=plugin_base.plugin_name,
        )
    except SimpleAdminTemplate.DoesNotExist:
        sat = SimpleAdminTemplate(
            template_id=template_id,
            plugin_type=plugin_base.plugin_type,
            plugin_name=plugin_base.plugin_name,
            automatic_created=True,
        )

    logger.debug(
        f"Registering template {plugin_base.plugin_type}/{plugin_base.plugin_name}/{template_id}"
    )

    sat.description = description
    sat.display_name = display_name
    sat.template = simpleadmin_template
    sat.update_method = update_method
    sat.modify_key = modify_key

    sat.save()

    return template_id


def scan_and_register_simpleadmin_template():
    existing_templates = {
        (sat.plugin_type, sat.plugin_name, sat.template_id): sat
        for sat in SimpleAdminTemplate.objects.filter(automatic_created=True)
    }

    for plugin_base in pluginhandler.get_all_plugins():
        if not hasattr(plugin_base, "simpleadmin_templates"):
            continue

        templates = plugin_base.simpleadmin_templates
        if templates == True:
            template_id = register_simpleadmin_template(plugin_base, templates)
            key = (plugin_base.plugin_type, plugin_base.plugin_name, template_id)
            existing_templates.pop(key, None)
        else:
            for template in templates:
                template_id = register_simpleadmin_template(plugin_base, template)
                key = (plugin_base.plugin_type, plugin_base.plugin_name, template_id)
                existing_templates.pop(key, None)

    for existing_template in existing_templates.values():
        logger.debug(
            f"It seems like {existing_template} is redundant, lets delete it if it is not in use"
        )
        if SimpleAdminPlugin.objects.filter(template=existing_template).exists():
            logger.debug(f"Template {existing_template} is in use, skipping")
            continue

        existing_template.delete()


class ServiceSimpleRouter(routers.SimpleRouter):
    def __init__(self, *args, **kwargs):
        self.service = kwargs.pop("service")
        super(ServiceSimpleRouter, self).__init__(*args, **kwargs)

    @property
    def routes(self):
        r = []
        for route in routers.SimpleRouter.routes:
            route = copy.copy(route)
            route.initkwargs["service"] = self.service
            r.append(route)
        return r


class AdminService(ServicePlugin):
    plugin_name = "admin"
    config_schema = Schema

    def __init__(self, config):
        scan_and_register_simpleadmin_template()
        super(AdminService, self).__init__(config)

    def get_urls(self):
        router = ServiceSimpleRouter(service=self)

        router.register("logs", LogModelView, basename="log")

        router.register("permissions", PermissionModelView, basename="permission")
        router.register("users", UserModelView, basename="user")
        router.register("plugins", PluginModelView, basename="plugin")
        router.register("pluginbases", PluginBaseListView, basename="pluginbase")
        router.register(
            "simpleadminplugins",
            SimpleAdminPluginModelView,
            basename="simpleadmin_plugin",
        )
        router.register(
            "simpleadmintemplates",
            SimpleAdminTemplateModelView,
            basename="simpleadmin_template",
        )

        router.register("schedules", ScheduleModelView)

        return [
            url("^$", ShowAdminUrlsView.as_view(urls=router.urls, service=self))
        ] + router.urls

    def unload(self):
        pass



from rest_framework import permissions, serializers, viewsets

from ....models import Log, LogMessage, Plugin
from .shared import ADMIN_RENDERER_CLASSES, ServiceAwareHyperlinkedIdentityField


class PluginSerializer(serializers.ModelSerializer):
    class Meta:
        model = Plugin
        fields = ("id", "name", "plugin_name", "plugin_type")


class LogMessageSerializer(serializers.ModelSerializer):
    class Meta:
        model = LogMessage
        fields = ("id", "datetime", "msg")

    class JSONAPIMeta:
        resource_name = "logmessage"


class LogSerializer(serializers.ModelSerializer):
    url = ServiceAwareHyperlinkedIdentityField(view_name="log-detail")
    username = serializers.CharField(source="user__username", allow_null=True)
    plugin = PluginSerializer(allow_null=True)

    included_serializers = {
        "log_messages": LogMessageSerializer,
        "plugin": PluginSerializer,
    }

    class Meta:
        model = Log
        fields = (
            "id",
            "username",
            "plugin",
            "action",
            "start_datetime",
            "end_datetime",
            "progress",
            "status",
            "url",
        )

    class JSONAPIMeta:
        resource_name = "log"


class LogDetailSerializer(LogSerializer):
    class Meta:
        model = Log
        fields = (
            "id",
            "username",
            "plugin",
            "action",
            "start_datetime",
            "end_datetime",
            "progress",
            "status",
            "log_messages",
        )

    class JSONAPIMeta:
        resource_name = "log"
        included_resources = ["log_messages", "plugin"]


class LogModelView(viewsets.ReadOnlyModelViewSet):
    queryset = Log.objects.all().order_by("-id")
    serializer_class = LogSerializer
    serializer_detail_class = LogDetailSerializer
    renderer_classes = ADMIN_RENDERER_CLASSES
    permission_classes = (permissions.IsAdminUser,)

    service = None

    def get_serializer_class(self):
        if self.action == "retrieve":
            return self.serializer_detail_class
        return super().get_serializer_class()



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'requirements.txt')) as f:
    requirements = f.read().strip().split('\n')


setup(
    name="unplugged",
    version="0.1.5",
    packages=find_packages(),
    package_data={
        '': [
            'unplugged/fixtures/unplugged-initial.json',
        ],
    },
    include_package_data=True,

    install_requires=requirements,
    author="Anders Jensen",
    author_email="johndoee@tidalstream.org",
    description="Django based plugin system",
    license="MIT",
    url="https://github.com/JohnDoee/unplugged",

)



import logging
import re

logger = logging.getLogger(__name__)


def register_urlpatterns(urls):
    logger.debug("Registering urls: %r" % (urls,))
    urlpatterns.extend(urls)


def unregister_urlpattern(pattern):
    logger.debug("Unregistering pattern: %s" % (pattern,))
    pattern = re.compile(pattern)
    for i, p in enumerate(urlpatterns):
        if pattern == p.pattern.regex:
            break
    else:
        return

    urlpatterns.pop(i)


def clear_urlpatterns():
    logger.debug("Clearing url patterns")
    urlpatterns[:] = []


urlpatterns = []



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "unplugged.services.staticurls"
    verbose_name = "StaticUrls Service"
    label = "services_staticurls"

    def ready(self):
        from .handler import StaticUrlsServicePlugin  # NOQA



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "unplugged.services.wamp"
    verbose_name = "WAMP Service"
    label = "services_wamp"

    def ready(self):
        from .handler import WAMPServicePlugin  # NOQA



import logging

from django.urls import include, path

from ...jsonapi import JSONAPIObject, JSONAPIRoot
from ...schema import Schema
from ...pluginhandler import pluginhandler
from ...plugins import ServicePlugin
from .views import StaticUrlsView

logger = logging.getLogger(__name__)


class StaticUrlsServicePlugin(ServicePlugin):
    plugin_name = "staticurls"
    config_schema = Schema
    plugin_url_root = None

    def get_urls(self):
        urls = [path("", StaticUrlsView.as_view(service=self))]
        root = JSONAPIRoot()
        for plugin_cls in pluginhandler.get_all_plugins():
            if not hasattr(plugin_cls, "get_static_urls"):
                continue
            static_urls = plugin_cls.get_static_urls()

            if not static_urls:
                continue

            logger.debug(
                f"Got {len(static_urls)} urls from {plugin_cls.plugin_type}/{plugin_cls.plugin_name}"
            )
            url = f"{plugin_cls.plugin_type}/{plugin_cls.plugin_name}/"
            urls.append(path(url, include(static_urls)))

            links = {"self": f"/{plugin_cls.name}/"}
            obj = JSONAPIObject(
                plugin_cls.plugin_type, plugin_cls.plugin_name, links=links
            )
            root.append(obj)

        self.plugin_url_root = root
        return urls

    def unload(self):
        pass



from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from wampyre.realm import realm_manager
from wampyre.transports.autowamp import ApplicationRunner

from .models import LogNotificationComponent, Plugin
from .scheduler import schedule_manager
from .signals import wamp_realm_created, wamp_realm_discarded


def handle_wampyre_callback(callback_type, realm):
    if callback_type == "create":
        wamp_realm_created.send(sender=realm)
    elif callback_type == "discard":
        wamp_realm_discarded.send(sender=realm)


def bootstrap_all():
    schedule_manager.start()
    Plugin.objects.bootstrap()

    content_type = ContentType.objects.get_for_model(Plugin)
    perm, _ = Permission.objects.get_or_create(
        content_type=content_type, codename="admin", defaults={"name": "Is admin"}
    )
    setattr(settings, "WAMP_LOG_TOPIC", f"group.{perm.id}.logs")

    realm_manager.register_callback(handle_wampyre_callback)

    ApplicationRunner(settings.WAMP_REALM).run(LogNotificationComponent)



import logging

from rest_framework import permissions
from rest_framework.response import Response
from rest_framework.views import APIView

from ...models import Plugin

logger = logging.getLogger(__name__)


class StaticUrlsView(APIView):
    service = None

    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request):
        return Response(self.service.plugin_url_root.serialize(request))



from django.conf.urls import url

from ...schema import Schema
from ...plugins import ServicePlugin
from .views import AuthView, LogoutView, UserView


class UserServicePlugin(ServicePlugin):
    plugin_name = "user"

    config_schema = Schema

    def get_urls(self):
        return [
            url("^/?$", UserView.as_view(service=self)),
            url("^auth/?$", AuthView.as_view(service=self)),
            url("^logout/?$", LogoutView.as_view(service=self)),
        ]

    def unload(self):
        pass



# Generated by Django 2.0.4 on 2019-07-15 19:44

import django.db.models.deletion
import jsonfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("unplugged", "0008_auto_20190715_1944")]

    operations = [
        migrations.CreateModel(
            name="SimpleAdminPlugin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True,
                        help_text="Slugified name, used for actual name (and be able to find the plugin again)",
                        max_length=200,
                        null=True,
                    ),
                ),
                (
                    "display_name",
                    models.CharField(
                        blank=True,
                        help_text="Human readable name",
                        max_length=200,
                        null=True,
                    ),
                ),
                ("config", jsonfield.fields.JSONField(blank=True, default=dict)),
                ("priority", models.SmallIntegerField(default=1)),
                (
                    "auto_created_plugins",
                    models.ManyToManyField(
                        blank=True,
                        related_name="auto_created_plugins",
                        to="unplugged.Plugin",
                    ),
                ),
                (
                    "plugin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="unplugged.Plugin",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Simple Admin Plugin",
                "db_table": "services_admin_simpleadmin_plugin",
                "ordering": ["priority"],
            },
        ),
        migrations.CreateModel(
            name="SimpleAdminTemplate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("display_name", models.CharField(max_length=200)),
                ("description", models.TextField(default="No description found")),
                ("template_id", models.CharField(max_length=100)),
                ("plugin_type", models.CharField(max_length=100)),
                ("plugin_name", models.CharField(max_length=100)),
                ("modify_key", models.CharField(blank=True, max_length=40, null=True)),
                ("template", jsonfield.fields.JSONField()),
                (
                    "update_method",
                    models.CharField(
                        choices=[("full", "Full"), ("modify_key", "Modify Key")],
                        max_length=100,
                    ),
                ),
                ("automatic_created", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name_plural": "Simple Admin Templates",
                "db_table": "services_admin_simpleadmin_template",
            },
        ),
        migrations.AlterUniqueTogether(
            name="simpleadmintemplate",
            unique_together={("template_id", "plugin_type", "plugin_name")},
        ),
        migrations.AddField(
            model_name="simpleadminplugin",
            name="template",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="services_admin.SimpleAdminTemplate",
            ),
        ),
    ]
